""" Looper version of NGS project model. """

import itertools
import os

from jsonschema import ValidationError
from pandas.core.common import flatten
from logging import getLogger

from peppy import SAMPLE_NAME_ATTR, OUTDIR_KEY, CONFIG_KEY, \
    Project as peppyProject
from eido import read_schema, PathAttrNotFoundError
from divvy import ComputingConfiguration
from ubiquerg import is_command_callable, expandpath

from .processed_project import populate_sample_paths, populate_project_paths
from .const import *
from .exceptions import *
from .utils import *
from .pipeline_interface import PipelineInterface

__all__ = ["Project"]

_LOGGER = getLogger(__name__)


class ProjectContext(object):
    """ Wrap a Project to provide protocol-specific Sample selection. """

    def __init__(self, prj, selector_attribute=None,
                 selector_include=None, selector_exclude=None):
        """ Project and what to include/exclude defines the context. """
        if not isinstance(selector_attribute, str):
            raise TypeError(
                "Name of attribute for sample selection isn't a string: {} "
                "({})".format(selector_attribute, type(selector_attribute)))
        self.prj = prj
        self.include = selector_include
        self.exclude = selector_exclude
        self.attribute = selector_attribute

    def __getattr__(self, item):
        """ Samples are context-specific; other requests are handled
        locally or dispatched to Project. """
        if item == "samples":
            return fetch_samples(prj=self.prj,
                                 selector_attribute=self.attribute,
                                 selector_include=self.include,
                                 selector_exclude=self.exclude)
        if item in ["prj", "include", "exclude"]:
            # Attributes requests that this context/wrapper handles
            return self.__dict__[item]
        else:
            # Dispatch attribute request to Project.
            return getattr(self.prj, item)

    def __getitem__(self, item):
        """ Provide the Mapping-like item access to the instance's Project. """
        return self.prj[item]

    def __enter__(self):
        """ References pass through this instance as needed, so the context
         provided is the instance itself. """
        return self

    def __repr__(self):
        return self.prj.__repr__()

    def __exit__(self, *args):
        """ Context teardown. """
        pass


class Project(peppyProject):
    """
    Looper-specific Project.

    :param str config_file: path to configuration file with data from
        which Project is to be built
    :param Iterable[str] amendments: name indicating amendment to use, optional
    :param str divcfg_path: path to an environment configuration YAML file
        specifying compute settings.
    :param bool permissive: Whether a error should be thrown if
        a sample input file(s) do not exist or cannot be open.
    :param str compute_env_file: Environment configuration YAML file specifying
        compute settings.
    """
    def __init__(self, config_file, amendments=None, divcfg_path=None,
                 runp=False, **kwargs):
        super(Project, self).__init__(config_file, amendments=amendments)
        setattr(self, EXTRA_KEY, dict())
        for attr_name in CLI_PROJ_ATTRS:
            if attr_name in kwargs:
                setattr(self[EXTRA_KEY], attr_name, kwargs[attr_name])
        if not runp:
            self._samples_by_interface = \
                self._samples_by_piface(self.piface_key)
            self._interfaces_by_sample = self._piface_by_samples()
        if FILE_CHECKS_KEY in self[EXTRA_KEY]:
            setattr(self, "file_checks", not self[EXTRA_KEY][FILE_CHECKS_KEY])
        if DRY_RUN_KEY in self[EXTRA_KEY]:
            setattr(self, DRY_RUN_KEY, self[EXTRA_KEY][DRY_RUN_KEY])
        self.dcc = None if divcfg_path is None else \
            ComputingConfiguration(filepath=divcfg_path)
        if hasattr(self, DRY_RUN_KEY) and not self[DRY_RUN_KEY]:
            _LOGGER.debug("Ensuring project directories exist")
            self.make_project_dirs()

    @property
    def piface_key(self):
        """
        Name of the pipeline interface attribute for this project

        :return str: name of the pipeline interface attribute
        """
        return self._extra_cli_or_cfg(PIFACE_KEY_SELECTOR) \
               or PIPELINE_INTERFACES_KEY

    @property
    def toggle_key(self):
        """
        Name of the toggle attribute for this project

        :return str: name of the toggle attribute
        """
        return self._extra_cli_or_cfg(TOGGLE_KEY_SELECTOR) or SAMPLE_TOGGLE_ATTR

    @property
    def selected_compute_package(self):
        """
        Compute package name specified in object constructor

        :return str: compute package name
        """
        return self._extra_cli_or_cfg(COMPUTE_PACKAGE_KEY)

    @property
    def cli_pifaces(self):
        """
        Collection of pipeline interface sources specified in object constructor

        :return list[str]: collection of pipeline interface sources
        """
        x = self._extra_cli_or_cfg(self.piface_key)
        return list(flatten([x] if not isinstance(x, list) else x))

    @property
    def output_dir(self):
        """
        Output directory for the project, specified in object constructor

        :return str: path to the output directory
        """
        return self._extra_cli_or_cfg(OUTDIR_KEY, strict=True)

    def _extra_cli_or_cfg(self, attr_name, strict=False):
        """
        Get attribute value provided in kwargs in object constructor of from
        looper section in the configuration file

        :param str attr_name: name of the attribute to get value for
        :param bool strict: whether a non-existent attribute is exceptional
        :raise MisconfigurationException: in strict mode, when no attribute
         found
        """
        try:
            result = getattr(self[EXTRA_KEY], attr_name)
        except (AttributeError, KeyError):
            return
        if result is not None:
            return result
        if CONFIG_KEY in self and LOOPER_KEY in self[CONFIG_KEY] \
                and attr_name in self[CONFIG_KEY][LOOPER_KEY]:
            return self[CONFIG_KEY][LOOPER_KEY][attr_name]
        else:
            if strict:
                raise MisconfigurationException(
                    "'{}' is missing. Provide it in the '{}' section of the "
                    "project configuration file".format(attr_name, LOOPER_KEY))
            return

    @property
    def results_folder(self):
        """
        Path to the results folder for the project

        :return str: path to the results folder in the output folder
        """
        return self._out_subdir_path(RESULTS_SUBDIR_KEY,
                                     default="results_pipeline")

    @property
    def submission_folder(self):
        """
        Path to the submission folder for the project

        :return str: path to the submission in the output folder
        """
        return self._out_subdir_path(SUBMISSION_SUBDIR_KEY,
                                     default="submission")

    def _out_subdir_path(self, key, default):
        """
        Create a system path relative to the project output directory.
        The values for the names of the subdirectories are sourced from
        kwargs passed to the object constructor.

        :param str key: name of the attribute mapped to the value of interest
        :param str default: if key not specified, a default to use
        :return str: path to the folder
        """
        return os.path.join(getattr(self, OUTDIR_KEY),
                            getattr(self[EXTRA_KEY], key) or default)

    def make_project_dirs(self):
        """
        Create project directory structure if it doesn't exist.
        """
        for folder_key in ["results_folder", "submission_folder"]:
            folder_path = getattr(self, folder_key)
            _LOGGER.debug("Ensuring project dir exists: '{}'".
                          format(folder_path))
            if not os.path.exists(folder_path):
                _LOGGER.debug("Attempting to create project folder: '{}'".
                              format(folder_path))
                try:
                    os.makedirs(folder_path)
                except OSError as e:
                    _LOGGER.warning("Could not create project folder: '{}'".
                                    format(str(e)))

    @property
    def project_pipeline_interface_sources(self):
        """
        Get a list of all valid project-level pipeline interface sources
        associated with this project. Sources that are file paths are expanded

        :return list[str]: collection of valid pipeline interface sources:
        """
        return [expandpath(src) for src in self.cli_pifaces] \
            if self.cli_pifaces is not None else []

    @property
    def project_pipeline_interfaces(self):
        """
        Flat list of all valid project-level interface objects associated
        with this Project

        Note that only valid pipeline interfaces will show up in the
        result (ones that exist on disk/remotely and validate successfully
        against the schema)

        :return list[looper.PipelineInterface]: list of pipeline interfaces
        """
        return [PipelineInterface(pi, pipeline_type="project")
                for pi in self.project_pipeline_interface_sources]

    @property
    def pipeline_interfaces(self):
        """
        Flat list of all valid interface objects associated with this Project

        Note that only valid pipeline interfaces will show up in the
        result (ones that exist on disk/remotely and validate successfully
        against the schema)

        :return list[looper.PipelineInterface]: list of pipeline interfaces
        """
        return [i for s in self._interfaces_by_sample.values() for i in s]

    @property
    def pipeline_interface_sources(self):
        """
        Get a list of all valid pipeline interface sources associated
        with this project. Sources that are file paths are expanded

        :return list[str]: collection of valid pipeline interface sources
        """
        return self._samples_by_interface.keys()

    # def _overwrite_sample_pifaces_with_cli(self, pifaces):
    #     """
    #     Overwrite sample pipeline interface sources with the provided ones
    #
    #     :param Iterable[str] | str | NoneType pifaces: collection of pipeline
    #         interface sources
    #     """
    #     _LOGGER.debug("CLI-specified pifaces: {}".format(pifaces))
    #     valid_pi = []
    #     if not pifaces:
    #         # No CLI-specified pipeline interface sources
    #         return
    #     if isinstance(pifaces, str):
    #         pifaces = [pifaces]
    #     for piface in pifaces:
    #         pi = expandpath(piface)
    #         try:
    #             PipelineInterface(pi, pipeline_type="sample")
    #         except Exception as e:
    #             _LOGGER.warning("Provided pipeline interface source ({}) is "
    #                             "invalid. Caught exception: {}".
    #                             format(pi, getattr(e, 'message', repr(e))))
    #         else:
    #             valid_pi.append(pi)
    #     [setattr(s, self.piface_key, valid_pi) for s in self.samples]
    #     if valid_pi:
    #         _LOGGER.info("Provided valid pipeline interface sources ({}) "
    #                      "set in all samples".format(", ".join(valid_pi)))

    def get_sample_piface(self, sample_name):
        """
        Get a list of pipeline interfaces associated with the specified sample.

        Note that only valid pipeline interfaces will show up in the
        result (ones that exist on disk/remotely and validate successfully
        against the schema)

        :param str sample_name: name of the sample to retrieve list of
            pipeline interfaces for
        :return list[looper.PipelineInterface]: collection of valid
            pipeline interfaces associated with selected sample
        """
        try:
            return self._interfaces_by_sample[sample_name]
        except KeyError:
            return None

    def build_submission_bundles(self, protocol, priority=True):
        """
        Create pipelines to submit for each sample of a particular protocol.

        With the argument (flag) to the priority parameter, there's control
        over whether to submit pipeline(s) from only one of the project's
        known pipeline locations with a match for the protocol, or whether to
        submit pipelines created from all locations with a match for the
        protocol.

        :param str protocol: name of the protocol/library for which to
            create pipeline(s)
        :param bool priority: to only submit pipeline(s) from the first of the
            pipelines location(s) (indicated in the project config file) that
            has a match for the given protocol; optional, default True
        :return Iterable[(PipelineInterface, type, str, str)]:
        :raises AssertionError: if there's a failure in the attempt to
            partition an interface's pipeline scripts into disjoint subsets of
            those already mapped and those not yet mapped
        """

        if not priority:
            raise NotImplementedError(
                "Currently, only prioritized protocol mapping is supported "
                "(i.e., pipeline interfaces collection is a prioritized list, "
                "so only the first interface with a protocol match is used.)")

        # Pull out the collection of interfaces (potentially one from each of
        # the locations indicated in the project configuration file) as a
        # sort of pool of information about possible ways in which to submit
        # pipeline(s) for sample(s) of the indicated protocol.
        pifaces = self.interfaces.get_pipeline_interface(protocol)
        if not pifaces:
            raise PipelineInterfaceConfigError(
                "No interfaces for protocol: {}".format(protocol))

        # coonvert to a list, in the future we might allow to match multiple
        pifaces = pifaces if isinstance(pifaces, str) else [pifaces]

        job_submission_bundles = []
        new_jobs = []

        _LOGGER.debug("Building pipelines matched by protocol: {}".
                      format(protocol))

        for pipe_iface in pifaces:
            # Determine how to reference the pipeline and where it is.
            path = pipe_iface["path"]
            if not (os.path.exists(path) or is_command_callable(path)):
                _LOGGER.warning("Missing pipeline script: {}".format(path))
                continue

            # Add this bundle to the collection of ones relevant for the
            # current PipelineInterface.
            new_jobs.append(pipe_iface)
            job_submission_bundles.append(new_jobs)
        return list(itertools.chain(*job_submission_bundles))

    @staticmethod
    def get_schemas(pifaces, schema_key=INPUT_SCHEMA_KEY):
        """
        Get the list of unique schema paths for a list of pipeline interfaces

        :param str | Iterable[str] pifaces: pipeline interfaces to search
            schemas for
        :param str schema_key: where to look for schemas in the piface
        :return Iterable[str]: unique list of schema file paths
        """
        if isinstance(pifaces, str):
            pifaces = [pifaces]
        schema_set = set()
        for piface in pifaces:
            schema_file = piface.get_pipeline_schemas(schema_key)
            if schema_file:
                schema_set.update([schema_file])
        return list(schema_set)

    def populate_pipeline_outputs(self, check_exist=False):
        """
        Populate project and sample output attributes based on output schemas
        that pipeline interfaces point to. Additionally, if requested,  check
        for the constructed paths existence on disk
        """
        for sample in self.samples:
            sample_piface = self.get_sample_piface(sample[SAMPLE_NAME_ATTR])
            if sample_piface:
                paths = self.get_schemas(sample_piface, OUTPUT_SCHEMA_KEY)
                for path in paths:
                    schema = read_schema(path)[-1]
                    try:
                        populate_project_paths(self, schema, check_exist)
                        populate_sample_paths(sample, schema, check_exist)
                    except PathAttrNotFoundError:
                        _LOGGER.error(
                            "Missing outputs of pipelines matched by protocol: "
                            "{}".format(sample.protocol)
                        )
                        raise

    def _piface_by_samples(self):
        """
        Create a mapping of all defined interfaces in this Project by samples.

        :return list[str]: a collection of pipeline interfaces keyed by
        sample name
        """
        pifaces_by_sample = {}
        for source, sample_names in self._samples_by_interface.items():
            for sample_name in sample_names:
                pifaces_by_sample.setdefault(sample_name, [])
                pifaces_by_sample[sample_name].\
                    append(PipelineInterface(source, pipeline_type="sample"))
        return pifaces_by_sample

    def _omit_from_repr(self, k, cls):
        """
        Exclude the interfaces from representation.

        :param str k: key of item to consider for omission
        :param type cls: placeholder to comply with superclass signature
        """
        return super(Project, self)._omit_from_repr(k, cls) or k == "interfaces"

    def _samples_by_piface(self, piface_key):
        """
        Create a collection of all samples with valid pipeline interfaces

        :param str piface_key: name of the attribute that holds pipeline
         interfaces
        :return list[str]: a collection of samples keyed by pipeline interface
            source
        """
        def _resolve_path(pth):
            """
            Expand provided path to the pipeline interface and make it absolute
            using project config path

            :param str pth: path, possibly including env vars and/or relative
            :return str: absolute path
            """
            pth = expandpath(pth)
            if not os.path.isabs(pth):
                pth = os.path.realpath(os.path.join(os.path.dirname(
                    self.config_file), pth))
                _LOGGER.debug("Relative path to pipeline interface source made "
                              "absolute: {}".format(pth))
            return pth
        samples_by_piface = {}
        msgs = set()
        for sample in self.samples:
            if piface_key in sample and sample[piface_key]:
                piface_srcs = sample[piface_key]
                if isinstance(piface_srcs, str):
                    piface_srcs = [piface_srcs]
                for source in piface_srcs:
                    source = _resolve_path(source)
                    try:
                        PipelineInterface(source, pipeline_type="sample")
                    except (ValidationError, IOError) as e:
                        msg = "Ignoring invalid pipeline interface source: " \
                              "{}. Caught exception: {}".\
                            format(source, getattr(e, 'message', repr(e)))
                        msgs.add(msg)
                        continue
                    else:
                        samples_by_piface.setdefault(source, set())
                        samples_by_piface[source].add(sample[SAMPLE_NAME_ATTR])
        for msg in msgs:
            _LOGGER.warning(msg)
        return samples_by_piface


def fetch_samples(prj, selector_attribute=None, selector_include=None,
                  selector_exclude=None):
    """
    Collect samples of particular protocol(s).

    Protocols can't be both positively selected for and negatively
    selected against. That is, it makes no sense and is not allowed to
    specify both selector_include and selector_exclude protocols. On the
    other hand, if
    neither is provided, all of the Project's Samples are returned.
    If selector_include is specified, Samples without a protocol will be
    excluded,
    but if selector_exclude is specified, protocol-less Samples will be
    included.

    :param Project prj: the Project with Samples to fetch
    :param str selector_attribute: name of attribute on which to base the
    fetch
    :param Iterable[str] | str selector_include: protocol(s) of interest;
        if specified, a Sample must
    :param Iterable[str] | str selector_exclude: protocol(s) to include
    :return list[Sample]: Collection of this Project's samples with
        protocol that either matches one of those in selector_include,
        or either
        lacks a protocol or does not match one of those in selector_exclude
    :raise TypeError: if both selector_include and selector_exclude
    protocols are
        specified; TypeError since it's basically providing two arguments
        when only one is accepted, so remain consistent with vanilla
        Python2;
        also possible if name of attribute for selection isn't a string
    """
    if selector_attribute is None or \
            (not selector_include and not selector_exclude):
        # Simple; keep all samples.  In this case, this function simply
        # offers a list rather than an iterator.
        return list(prj.samples)

    if not isinstance(selector_attribute, str):
        raise TypeError(
            "Name for attribute on which to base selection isn't string: "
            "{} "
            "({})".format(selector_attribute, type(selector_attribute)))

    # At least one of the samples has to have the specified attribute
    if prj.samples and not any(
            [hasattr(s, selector_attribute) for s in prj.samples]):
        raise AttributeError(
            "The Project samples do not have the attribute '{attr}'".
                format(attr=selector_attribute))

    # Intersection between selector_include and selector_exclude is
    # nonsense user error.
    if selector_include and selector_exclude:
        raise TypeError(
            "Specify only selector_include or selector_exclude parameter, "
            "not both.")

    # Ensure that we're working with sets.
    def make_set(items):
        if isinstance(items, str):
            items = [items]
        return items

    # Use the attr check here rather than exception block in case the
    # hypothetical AttributeError would occur; we want such
    # an exception to arise, not to catch it as if the Sample lacks
    # "protocol"
    if not selector_include:
        # Loose; keep all samples not in the selector_exclude.
        def keep(s):
            return not hasattr(s, selector_attribute) \
                   or getattr(s, selector_attribute) \
                   not in make_set(selector_exclude)
    else:
        # Strict; keep only samples in the selector_include.
        def keep(s):
            return hasattr(s, selector_attribute) \
                   and getattr(s, selector_attribute) \
                   in make_set(selector_include)

    return list(filter(keep, prj.samples))



""" Generate HTML reports """

import os
import glob
import pandas as _pd
import logging
import jinja2
import re
import sys
from warnings import warn
from datetime import timedelta
from ._version import __version__ as v
from .const import *
from .processed_project import get_project_outputs
from .utils import get_file_for_project
from peppy.const import *
from eido import read_schema
from copy import copy as cp
_LOGGER = logging.getLogger("looper")


class HTMLReportBuilder(object):
    """ Generate HTML summary report for project/samples """

    def __init__(self, prj):
        """
        The Project defines the instance.

        :param Project prj: Project with which to work/operate on
        """
        super(HTMLReportBuilder, self).__init__()
        self.prj = prj
        self.j_env = get_jinja_env()
        self.reports_dir = get_file_for_project(self.prj, "reports")
        self.index_html_path = get_file_for_project(self.prj, "summary.html")
        self.index_html_filename = os.path.basename(self.index_html_path)
        self._outdir = self.prj.output_dir
        _LOGGER.debug("Reports dir: {}".format(self.reports_dir))

    def __call__(self, objs, stats, columns):
        """ Do the work of the subcommand/program. """
        # Generate HTML report
        navbar = self.create_navbar(self.create_navbar_links(
            objs=objs, stats=stats,
            wd=self._outdir),
            self.index_html_filename)
        navbar_reports = self.create_navbar(
            self.create_navbar_links(
                objs=objs, stats=stats, wd=self.reports_dir),
            os.path.join("..", self.index_html_filename))
        index_html_path = self.create_index_html(
            objs, stats, columns, footer=self.create_footer(),
            navbar=navbar, navbar_reports=navbar_reports)
        return index_html_path

    def create_object_parent_html(self, objs, navbar, footer):
        """
        Generates a page listing all the project objects with links
        to individual object pages

        :param pandas.DataFrame objs: project level dataframe containing any reported objects for all samples
        :param str navbar: HTML to be included as the navbar in the main summary page
        :param str footer: HTML to be included as the footer
        :return str: Rendered parent objects HTML file
        """
        object_parent_path = os.path.join(self.reports_dir, "objects.html")

        if not os.path.exists(os.path.dirname(object_parent_path)):
            os.makedirs(os.path.dirname(object_parent_path))
        pages = list()
        labels = list()
        if not objs.empty:
            for key in objs['key'].drop_duplicates().sort_values():
                page_name = key + ".html"
                page_path = os.path.join(self.reports_dir, page_name.replace(' ', '_').lower())
                page_relpath = os.path.relpath(page_path, self.reports_dir)
                pages.append(page_relpath)
                labels.append(key)

        template_vars = dict(navbar=navbar, footer=footer, labels=labels, pages=pages, header="Objects")
        return render_jinja_template("navbar_list_parent.html", self.j_env, template_vars)

    def create_sample_parent_html(self, navbar, footer):
        """
        Generates a page listing all the project samples with links
        to individual sample pages
        :param str navbar: HTML to be included as the navbar in the main summary page
        :param str footer: HTML to be included as the footer
        :return str: Rendered parent samples HTML file
        """
        sample_parent_path = os.path.join(self.reports_dir, "samples.html")

        if not os.path.exists(os.path.dirname(sample_parent_path)):
            os.makedirs(os.path.dirname(sample_parent_path))
        pages = list()
        labels = list()
        for sample in self.prj.samples:
            sample_name = str(sample.sample_name)
            sample_dir = os.path.join(
                self.prj.results_folder, sample_name)

            # Confirm sample directory exists, then build page
            if os.path.exists(sample_dir):
                page_name = sample_name + ".html"
                page_path = os.path.join(self.reports_dir, page_name.replace(' ', '_').lower())
                page_relpath = os.path.relpath(page_path, self.reports_dir)
                pages.append(page_relpath)
                labels.append(sample_name)

        template_vars = dict(navbar=navbar, footer=footer, labels=labels, pages=pages, header="Samples")
        return render_jinja_template("navbar_list_parent.html", self.j_env, template_vars)

    def create_navbar(self, navbar_links, index_html_relpath):
        """
        Creates the navbar using the privided links

        :param str navbar_links: HTML list of links to be inserted into a navbar
        :return str: navbar HTML
        """
        template_vars = dict(navbar_links=navbar_links, index_html=index_html_relpath)
        return render_jinja_template("navbar.html", self.j_env, template_vars)

    def create_footer(self):
        """
        Renders the footer from the templates directory

        :return str: footer HTML
        """
        return render_jinja_template("footer.html", self.j_env, dict(version=v))

    def create_navbar_links(self, objs, stats, wd=None, context=None, include_status=True):
        """
        Return a string containing the navbar prebuilt html.

        Generates links to each page relative to the directory of interest (wd arg) or uses the provided context to
        create the paths (context arg)

        :param pandas.DataFrame objs: project results dataframe containing
            object data
        :param list stats[dict] stats: a summary file of pipeline statistics for each
            analyzed sample
        :param path wd: the working directory of the current HTML page being generated, enables navbar links
            relative to page
        :param list[str] context: the context the links will be used in.
            The sequence of directories to be prepended to the HTML file in the resulting navbar
        :param bool include_status: whether the status link should be included in the links set
        :return str: navbar links as HTML-formatted string
        """
        if wd is None and context is None:
            raise ValueError("Either 'wd' (path the links should be relative to) or 'context'"
                             " (the context for the links) has to be provided.")
        status_relpath = _make_relpath(file_name=os.path.join(self.reports_dir, "status.html"), wd=wd, context=context)
        objects_relpath = _make_relpath(file_name=os.path.join(self.reports_dir, "objects.html"), wd=wd, context=context)
        samples_relpath = _make_relpath(file_name=os.path.join(self.reports_dir, "samples.html"), wd=wd, context=context)
        dropdown_keys_objects = None
        dropdown_relpaths_objects = None
        dropdown_relpaths_samples = None
        sample_names = None
        if objs is not None and not objs.dropna().empty:
            # If the number of objects is 20 or less, use a drop-down menu
            if len(objs['key'].drop_duplicates()) <= 20:
                dropdown_relpaths_objects, dropdown_keys_objects = \
                    _get_navbar_dropdown_data_objects(objs=objs, wd=wd, context=context, reports_dir=self.reports_dir)
            else:
                dropdown_relpaths_objects = objects_relpath
        if stats:
            if len(stats) <= 20:
                dropdown_relpaths_samples, sample_names = \
                    _get_navbar_dropdown_data_samples(stats=stats, wd=wd, context=context, reports_dir=self.reports_dir)
            else:
                # Create a menu link to the samples parent page
                dropdown_relpaths_samples = samples_relpath
        status_page_name = "Status" if include_status else None
        template_vars = dict(status_html_page=status_relpath, status_page_name=status_page_name,
                             dropdown_keys_objects=dropdown_keys_objects, objects_page_name="Objects",
                             samples_page_name="Samples", objects_html_page=dropdown_relpaths_objects,
                             samples_html_page=dropdown_relpaths_samples, menu_name_objects="Objects",
                             menu_name_samples="Samples", sample_names=sample_names, all_samples=samples_relpath,
                             all_objects=objects_relpath)
        return render_jinja_template("navbar_links.html", self.j_env, template_vars)

    def create_object_html(self, single_object, navbar, footer):
        """
        Generates a page for an individual object type with all of its
        plots from each sample

        :param pandas.DataFrame single_object: contains reference
            information for an individual object type for all samples
        :param pandas.DataFrame objs: project level dataframe
            containing any reported objects for all samples
        :param str navbar: HTML to be included as the navbar in the main summary page
        :param str footer: HTML to be included as the footer
        """

        # Generate object filename
        for key in single_object['key'].drop_duplicates().sort_values():
            # even though it's always one element, loop to extract the data
            current_name = str(key)
            filename = current_name + ".html"
        html_page_path = os.path.join(self.reports_dir, filename.replace(' ', '_').lower())

        if not os.path.exists(os.path.dirname(html_page_path)):
            os.makedirs(os.path.dirname(html_page_path))

        links = []
        figures = []
        warnings = []
        for i, row in single_object.iterrows():
            # Set the PATH to a page for the sample. Catch any errors.
            try:
                object_path = os.path.join(self.prj.results_folder, row['sample_name'], row['filename'])
                object_relpath = os.path.relpath(object_path, self.reports_dir)
            except AttributeError:
                err_msg = ("Sample: {} | " + "Missing valid object path for: {}")
                # Report the sample that fails, if that information exists
                if str(row['sample_name']) and str(row['filename']):
                    _LOGGER.warning(err_msg.format(row['sample_name'], row['filename']))
                else:
                    _LOGGER.warning(err_msg.format("Unknown sample"))
                object_relpath = ""

            # Set the PATH to the image/file. Catch any errors.
            # Check if the object is an HTML document

            if not str(row['anchor_image']).lower().endswith(IMAGE_EXTS):
                image_path = object_path
            else:
                try:
                    image_path = os.path.join(self.prj.results_folder, row['sample_name'], row['anchor_image'])
                except AttributeError:
                    _LOGGER.warning(str(row))
                    err_msg = ("Sample: {} | " + "Missing valid image path for: {}")
                    # Report the sample that fails, if that information exists
                    if str(row['sample_name']) and str(row['filename']):
                        _LOGGER.warning(err_msg.format(row['sample_name'], row['filename']))
                    else:
                        _LOGGER.warning(err_msg.format("Unknown", "Unknown"))
                    image_path = ""
            # Check for the presence of both the file and thumbnail
            if os.path.isfile(image_path) and os.path.isfile(object_path):
                image_relpath = os.path.relpath(image_path, self.reports_dir)
                # If the object has a valid image, use it!
                _LOGGER.debug("Checking image path: {}".format(image_path))
                if str(image_path).lower().endswith(IMAGE_EXTS):
                    figures.append([object_relpath, str(row['sample_name']), image_relpath])
                # Or if that "image" is not an image, treat it as a link
                elif not str(image_path).lower().endswith(IMAGE_EXTS):
                    _LOGGER.debug("Got link")
                    links.append([str(row['sample_name']), image_relpath])
            else:
                warnings.append(str(row['filename']))

        if warnings:
            _LOGGER.warning("create_object_html: " +
                            filename.replace(' ', '_').lower() + " references nonexistent object files")
            _LOGGER.debug(filename.replace(' ', '_').lower() +
                          " nonexistent files: " + ','.join(str(x) for x in warnings))
        template_vars = dict(navbar=navbar, footer=footer, name=current_name, figures=figures, links=links)
        save_html(html_page_path, render_jinja_template("object.html", self.j_env, args=template_vars))

    def create_sample_html(self, objs, sample_name, sample_stats, navbar, footer):
        """
        Produce an HTML page containing all of a sample's objects
        and the sample summary statistics

        :param pandas.DataFrame objs: project level dataframe containing
            any reported objects for all samples
        :param str sample_name: the name of the current sample
        :param dict sample_stats: pipeline run statistics for the current sample
        :param str navbar: HTML to be included as the navbar in the main summary page
        :param str footer: HTML to be included as the footer
        :return str: path to the produced HTML page
        """
        html_filename = sample_name + ".html"
        html_page = os.path.join(self.reports_dir, html_filename.replace(' ', '_').lower())
        sample_page_relpath = os.path.relpath(html_page, self._outdir)
        single_sample = _pd.DataFrame() if objs.empty else objs[objs['sample_name'] == sample_name]
        if not os.path.exists(os.path.dirname(html_page)):
            os.makedirs(os.path.dirname(html_page))
        sample_dir = os.path.join(self.prj.results_folder, sample_name)
        if os.path.exists(sample_dir):
            if single_sample.empty:
                # When there is no objects.tsv file, search for the
                # presence of log, profile, and command files
                log_name = _match_file_for_sample(sample_name, 'log.md', self.prj.results_folder)
                profile_name = _match_file_for_sample(sample_name, 'profile.tsv', self.prj.results_folder)
                command_name = _match_file_for_sample(sample_name, 'commands.sh', self.prj.results_folder)
            else:
                log_name = str(single_sample.iloc[0]['annotation']) + "_log.md"
                profile_name = str(single_sample.iloc[0]['annotation']) + "_profile.tsv"
                command_name = str(single_sample.iloc[0]['annotation']) + "_commands.sh"
            stats_name = "stats.tsv"
            flag = _get_flags(sample_dir)
            # get links to the files
            stats_file_path = _get_relpath_to_file(
                stats_name, sample_name, self.prj.results_folder, self.reports_dir)
            profile_file_path = _get_relpath_to_file(
                profile_name, sample_name, self.prj.results_folder, self.reports_dir)
            commands_file_path = _get_relpath_to_file(
                command_name, sample_name, self.prj.results_folder, self.reports_dir)
            log_file_path = _get_relpath_to_file(
                log_name, sample_name, self.prj.results_folder, self.reports_dir)
            if not flag:
                button_class = "btn btn-secondary"
                flag = "Missing"
            elif len(flag) > 1:
                button_class = "btn btn-secondary"
                flag = "Multiple"
            else:
                flag = flag[0]
                try:
                    flag_dict = BUTTON_APPEARANCE_BY_FLAG[flag]
                except KeyError:
                    button_class = "btn btn-secondary"
                    flag = "Unknown"
                else:
                    button_class = flag_dict["button_class"]
                    flag = flag_dict["flag"]
        links = []
        figures = []
        warnings = []
        if not single_sample.empty:
            for sample_name in single_sample['sample_name'].drop_duplicates().sort_values():
                o = single_sample[single_sample['sample_name'] == sample_name]
                for i, row in o.iterrows():
                    try:
                        # Image thumbnails are optional
                        # This references to "image" should really
                        # be "thumbnail"
                        image_path = os.path.join(
                            self.prj.results_folder,
                            sample_name, row['anchor_image'])
                        image_relpath = os.path.relpath(image_path, self.reports_dir)
                    except (AttributeError, TypeError):
                        image_path = ""
                        image_relpath = ""

                    # These references to "page" should really be
                    # "object", because they can be anything.
                    page_path = os.path.join(
                        self.prj.results_folder,
                        sample_name, row['filename'])
                    page_relpath = os.path.relpath(page_path, self.reports_dir)
                    # If the object has a thumbnail image, add as a figure
                    if os.path.isfile(image_path) and os.path.isfile(page_path):
                        # If the object has a valid image, add as a figure
                        if str(image_path).lower().endswith(('.png', '.jpg', '.jpeg', '.svg', '.gif')):
                            figures.append([page_relpath, str(row['key']), image_relpath])
                        # Otherwise treat as a link
                        elif os.path.isfile(page_path):
                            links.append([str(row['key']), page_relpath])
                        # If neither, there is no object by that name
                        else:
                            warnings.append(str(row['filename']))
                    # If no thumbnail image, it's just a link
                    elif os.path.isfile(page_path):
                        links.append([str(row['key']), page_relpath])
                    # If no file present, there is no object by that name
                    else:
                        warnings.append(str(row['filename']))
        else:
            # Sample was not run through the pipeline
            _LOGGER.warning("{} is not present in {}".format(
                sample_name, self.prj.results_folder))

        template_vars = dict(navbar=navbar, footer=footer, sample_name=sample_name, stats_file_path=stats_file_path,
                             profile_file_path=profile_file_path, commands_file_path=commands_file_path,
                             log_file_path=log_file_path, button_class=button_class, sample_stats=sample_stats,
                             flag=flag, links=links, figures=figures)
        save_html(html_page, render_jinja_template("sample.html", self.j_env, template_vars))
        return sample_page_relpath

    def create_status_html(self, status_table, navbar, footer):
        """
        Generates a page listing all the samples, their run status, their
        log file, and the total runtime if completed.

        :param pandas.DataFrame objs: project level dataframe containing any reported objects for all samples
        :param str navbar: HTML to be included as the navbar in the main summary page
        :param str footer: HTML to be included as the footer
        :return str: rendered status HTML file
        """
        _LOGGER.debug("Building status page...")
        template_vars = dict(status_table=status_table, navbar=navbar,
                             footer=footer)
        return render_jinja_template("status.html", self.j_env, template_vars)

    def create_project_objects(self):
        """
        Render available project level outputs defined in the
        pipeline output schemas
        """
        _LOGGER.debug("Building project objects section...")
        figures = []
        links = []
        warnings = []
        # For each protocol report the project summarizers' results
        self.prj.populate_pipeline_outputs()
        ifaces = self.prj.project_pipeline_interfaces
        # Check the interface files for summarizers
        for iface in ifaces:
            schema_paths = \
                iface.get_pipeline_schemas(OUTPUT_SCHEMA_KEY)
            if schema_paths is not None:
                if isinstance(schema_paths, str):
                    schema_paths = [schema_paths]
                for output_schema_path in schema_paths:
                    results = get_project_outputs(
                        self.prj, read_schema(output_schema_path))
                    for name, result in results.items():
                        title = str(result.setdefault('title', "No caption"))
                        result_type = str(result['type'])
                        result_file = str(result['path'])
                        result_img = \
                            str(result.setdefault('thumbnail_path', None))
                        if result_img and not os.path.isabs(result_file):
                            result_img = os.path.join(
                                self._outdir, result_img)
                        if not os.path.isabs(result_file):
                            result_file = os.path.join(
                                self._outdir, result_file)
                        _LOGGER.debug("Looking for project file: {}".
                                      format(result_file))
                        # Confirm the file itself was produced
                        if glob.glob(result_file):
                            file_path = str(glob.glob(result_file)[0])
                            file_relpath = \
                                os.path.relpath(file_path, self._outdir)
                            if result_type == "image":
                                # Add as a figure, find thumbnail
                                search = os.path.join(self._outdir, result_img)
                                if glob.glob(search):
                                    img_path = str(glob.glob(search)[0])
                                    img_relpath = \
                                        os.path.relpath(img_path, self._outdir)
                                    figures.append(
                                        [file_relpath, title, img_relpath])
                            # add as a link otherwise
                            # TODO: add more fine-grained type support?
                            #  not just image and link
                            else:
                                links.append([title, file_relpath])
                        else:
                            warnings.append("{} ({})".format(title,
                                                             result_file))
            else:
                _LOGGER.debug("No project-level outputs defined in "
                              "schema: {}".format(schema_paths))
        if warnings:
            _LOGGER.warning("Not found: {}".
                            format([str(x) for x in warnings]))
        _LOGGER.debug("collected project-level figures: {}".format(figures))
        _LOGGER.debug("collected project-level links: {}".format(links))
        template_vars = dict(figures=figures, links=links)
        return render_jinja_template("project_object.html", self.j_env,
                                     template_vars)

    def create_index_html(self, objs, stats, col_names, navbar, footer, navbar_reports=None):
        """
        Generate an index.html style project home page w/ sample summary
        statistics

        :param pandas.DataFrame objs: project level dataframe containing
            any reported objects for all samples
        :param list[dict] stats: a summary file of pipeline statistics for each
            analyzed sample
        :param list col_names: all unique column names used in the stats file
        :param str navbar: HTML to be included as the navbar in the main summary page
        :param str footer: HTML to be included as the footer
        :param str navbar_reports: HTML to be included as the navbar for pages in the reports directory
        """
        # set default encoding when running in python2
        if sys.version[0] == '2':
            from importlib import reload
            reload(sys)
            sys.setdefaultencoding("utf-8")
        _LOGGER.debug("Building index page...")
        # copy the columns names and remove the sample_name one, since it will be processed differently
        cols = cp(col_names)
        cols.remove("sample_name")
        if navbar_reports is None:
            navbar_reports = navbar
        if not objs.dropna().empty:
            objs.drop_duplicates(keep='last', inplace=True)
        # Generate parent index.html page path
        index_html_path = get_file_for_project(self.prj, "summary.html")

        # Add stats_summary.tsv button link
        stats_file_name = os.path.join(self._outdir, self.prj.name)
        if hasattr(self.prj, "subproject") and self.prj.subproject:
            stats_file_name += '_' + self.prj.subproject
        stats_file_name += '_stats_summary.tsv'
        stats_file_path = os.path.relpath(stats_file_name, self._outdir)
        # Add stats summary table to index page and produce individual
        # sample pages
        if os.path.isfile(stats_file_name):
            # Produce table rows
            table_row_data = []
            samples_cols_missing = []
            _LOGGER.debug(" * Creating sample pages...")
            for row in stats:
                table_cell_data = []
                sample_name = row["sample_name"]
                sample_page = self.create_sample_html(objs, sample_name, row, navbar_reports, footer)
                # treat sample_name column differently - provide a link to the sample page
                table_cell_data.append([sample_page, sample_name])
                # for each column read the data from the stats
                for c in cols:
                    try:
                        table_cell_data.append(str(row[c]))
                    except KeyError:
                        table_cell_data.append("NA")
                        samples_cols_missing.append(sample_name)
                table_row_data.append(table_cell_data)
            _LOGGER.debug("Samples with missing columns: {}".format(set(samples_cols_missing)))
        else:
            _LOGGER.warning("No stats file '%s'", stats_file_name)

        # Create parent samples page with links to each sample
        save_html(os.path.join(self.reports_dir, "samples.html"), self.create_sample_parent_html(navbar_reports, footer))
        _LOGGER.debug(" * Creating object pages...")
        # Create objects pages
        if not objs.dropna().empty:
            for key in objs['key'].drop_duplicates().sort_values():
                single_object = objs[objs['key'] == key]
                self.create_object_html(single_object, navbar_reports, footer)

        # Create parent objects page with links to each object type
        save_html(os.path.join(self.reports_dir, "objects.html"),
                  self.create_object_parent_html(objs, navbar_reports, footer))
        # Create status page with each sample's status listed
        save_html(os.path.join(self.reports_dir, "status.html"),
                  self.create_status_html(create_status_table(self.prj), navbar_reports, footer))
        # Add project level objects
        project_objects = self.create_project_objects()
        # Complete and close HTML file
        template_vars = dict(project_name=self.prj.name, stats_json=_read_tsv_to_json(stats_file_name),
                             navbar=navbar, footer=footer, stats_file_path=stats_file_path,
                             project_objects=project_objects, columns=col_names, table_row_data=table_row_data)
        save_html(index_html_path, render_jinja_template("index.html", self.j_env, template_vars))
        return index_html_path


def render_jinja_template(name, jinja_env, args=dict()):
    """
    Render template in the specified jinja environment using the provided args

    :param str name: name of the template
    :param dict args: arguments to pass to the template
    :param jinja2.Environment jinja_env: the initialized environment to use in this the looper HTML reports context
    :return str: rendered template
    """
    assert isinstance(args, dict), "args has to be a dict"
    template = jinja_env.get_template(name)
    return template.render(**args)


def save_html(path, template):
    """
    Save rendered template as an HTML file

    :param str path: the desired location for the file to be produced
    :param str template: the template or just string
    """
    if not os.path.exists(os.path.dirname(path)):
        os.makedirs(os.path.dirname(path))
    try:
        with open(path, 'w') as f:
            f.write(template)
    except IOError:
        _LOGGER.error("Could not write the HTML file: {}".format(path))


def get_jinja_env(templates_dirname=None):
    """
    Create jinja environment with the provided path to the templates directory

    :param str templates_dirname: path to the templates directory
    :return jinja2.Environment: jinja environment
    """
    if templates_dirname is None:
        file_dir = os.path.dirname(os.path.realpath(__file__))
        templates_dirname = os.path.join(file_dir, TEMPLATES_DIRNAME)
    _LOGGER.debug("Using templates dir: " + templates_dirname)
    return jinja2.Environment(loader=jinja2.FileSystemLoader(templates_dirname))


def _get_flags(sample_dir):
    """
    Get the flag(s) present in the directory

    :param str sample_dir: path to the directory to be searched for flags
    :return list: flags found in the dir
    """
    assert os.path.exists(sample_dir), "The provided path ('{}') does not exist".format(sample_dir)
    flag_files = glob.glob(os.path.join(sample_dir, '*.flag'))
    if len(flag_files) > 1:
        _LOGGER.warning("Multiple flag files ({files_count}) found in sample dir '{sample_dir}'".
                        format(files_count=len(flag_files), sample_dir=sample_dir))
    if len(flag_files) == 0:
        _LOGGER.warning("No flag files found in sample dir '{sample_dir}'".format(sample_dir=sample_dir))
    return [re.search(r'\_([a-z]+)\.flag$', os.path.basename(f)).groups()[0] for f in flag_files]


def _match_file_for_sample(sample_name, appendix, location, full_path=False):
    """
    Safely looks for files matching the appendix in the specified location for the sample

    :param str sample_name: name of the sample that the file name should be found for
    :param str appendix: the ending  specific for the file
    :param str location: where to look for the file
    :param bool full_path: whether to return full path
    :return str: the name of the matched file
    """
    regex = "*" + appendix
    search_pattern = os.path.join(location, sample_name, regex)
    matches = glob.glob(search_pattern)
    if len(matches) < 1:
        return None
    elif len(matches) > 1:
        _LOGGER.warning("matched mutiple files for '{}'. Returning the first one".format(search_pattern))
    return matches[0] if full_path else os.path.basename(matches[0])


def _get_relpath_to_file(file_name, sample_name, location, relative_to):
    """
    Safely gets the relative path for the file for the specified sample

    :param str file_name: name of the file
    :param str sample_name: name of the sample that the file path should be found for
    :param str location: where to look for the file
    :param str relative_to: path the result path should be relative to
    :return str: a path to the file
    """
    abs_file_path = os.path.join(location, sample_name, file_name)
    rel_file_path = os.path.relpath(abs_file_path, relative_to)
    if file_name is None or not os.path.exists(abs_file_path):
        return None
    return rel_file_path


def _make_relpath(file_name, wd, context=None):
    """
    Create a path relative to the context. This function introduces the flexibility to the navbar links creation,
    which the can be used outside of the native looper summary pages.

    :param str file_name: the path to make relative
    :param str wd: the dir the path should be relative to
    :param list[str] context: the context the links will be used in.
            The sequence of directories to be prepended to the HTML file in the resulting navbar
    :return str: relative path
    """
    relpath = os.path.relpath(file_name, wd)
    return relpath if not context else os.path.join(os.path.join(*context), relpath)


def _get_navbar_dropdown_data_objects(objs, wd, context, reports_dir):
    if objs is None:
        return None, None
    relpaths = []
    df_keys = objs['key'].drop_duplicates().sort_values()
    for key in df_keys:
        page_name = os.path.join(reports_dir, (key + ".html").replace(' ', '_').lower())
        relpaths.append(_make_relpath(page_name, wd, context))
    return relpaths, df_keys


def _get_navbar_dropdown_data_samples(stats, wd, context, reports_dir):
    if stats is None:
        return None, None
    relpaths = []
    sample_names = []
    for sample in stats:
        for entry, val in sample.items():
            if entry == "sample_name":
                sample_name = str(val)
                page_name = os.path.join(reports_dir, (sample_name + ".html").replace(' ', '_').lower())
                relpaths.append(_make_relpath(page_name, wd, context))
                sample_names.append(sample_name)
                break
            else:
                _LOGGER.warning("Could not determine sample name in stats.tsv")
    return relpaths, sample_names


def _read_csv_encodings(path, encodings=["utf-8", "ascii"], **kwargs):
    """
    Try to read file with the provided encodings

    :param str path: path to file
    :param list encodings: list of encodings to try
    """
    idx = 0
    while idx < len(encodings):
        e = encodings[idx]
        try:
            t = _pd.read_csv(path, encoding=e, **kwargs)
            return t
        except UnicodeDecodeError:
            pass
        idx = idx + 1
    _LOGGER.warning("Could not read the log file '{p}' with encodings '{enc}'".format(p=path, enc=encodings))


def _get_from_log(log_path, regex):
    """
    Get the value for the matched key from log file

    :param str log_path: path to the log file
    :param str regex: matching str. Should be formatted as follows: r'(phrase to match)'
    :return str: matched and striped string
    :raises IOError: when the file is not found in the provided path
    """
    if not os.path.exists(log_path):
        raise IOError("Can't read the log file '{}'. Not found".format(log_path))
    log = _read_csv_encodings(log_path, header=None, names=['data'])
    if log is None:
        _LOGGER.warning("'{r}' was not read from log".format(r=regex))
        return None
    # match regex, get row(s) that matched the regex
    log_row = log.iloc[:, 0].str.extractall(regex)
    # not matches? return None
    if log_row.empty:
        return None
    if log_row.size > 1:
        _LOGGER.warning("When parsing '{lp}', more than one values matched with: {r}. Returning first.".format(lp=log_path, r=regex))
    # split the matched line by first colon return stripped data.
    # This way both mem values (e.g 1.1GB) and time values (e.g 1:10:10) will work.
    val = log.iloc[log_row.index[0][0]].str.split(":", 1, expand=True)[1][0].strip()
    return val


def _read_tsv_to_json(path):
    """
    Read a tsv file to a JSON formatted string

    :param path: to file path
    :return str: JSON formatted string
    """
    assert os.path.exists(path), "The file '{}' does not exist".format(path)
    _LOGGER.debug("Reading TSV from '{}'".format(path))
    df = _pd.read_csv(path, sep="\t", index_col=False, header=None)
    return df.to_json()


def uniqify(seq):
    """ Fast way to uniqify while preserving input order. """
    # http://stackoverflow.com/questions/480214/
    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]


def create_status_table(prj, final=True):
    """
    Creates status table, the core of the status page.
    It is abstracted into a function so that it can be used in other software
    packages. It can produce a table of two types. With links to the
    samples/log files and without. The one without can be used to render HTMLs
     for on-th-fly job status inspection.

    :param looper.Project prj: project to create the status table for
    :param bool final: if the status table is created for a finalized looper
        run. In such a case, links to samples and log files will be provided
    :return str: rendered status HTML file
    """
    status_warning = False
    sample_warning = []
    log_paths = []
    log_link_names = []
    sample_paths = []
    sample_link_names = []
    flags = []
    row_classes = []
    times = []
    mems = []
    for sample in prj.samples:
        sample_name = str(sample.sample_name)
        sample_dir = os.path.join(
            prj.results_folder, sample_name)

        # Confirm sample directory exists, then build page
        if os.path.exists(sample_dir):
            # Grab the status flag for the current sample
            flag = _get_flags(sample_dir)
            if not flag:
                button_class = "table-secondary"
                flag = "Missing"
            elif len(flag) > 1:
                button_class = "table-secondary"
                flag = "Multiple"
            else:
                flag = flag[0]
                try:
                    flag_dict = TABLE_APPEARANCE_BY_FLAG[flag]
                except KeyError:
                    button_class = "table-secondary"
                    flag = "Unknown"
                else:
                    button_class = flag_dict["button_class"]
                    flag = flag_dict["flag"]
            row_classes.append(button_class)
            # get first column data (sample name/link)
            page_name = sample_name + ".html"
            page_path = os.path.join(get_file_for_project(prj, "reports"),
                                     page_name.replace(' ', '_').lower())
            page_relpath = os.path.relpath(page_path,
                                           get_file_for_project(prj, "reports"))
            sample_paths.append(page_relpath)
            sample_link_names.append(sample_name)
            # get second column data (status/flag)
            flags.append(flag)
            # get third column data (log file/link)
            log_name = _match_file_for_sample(sample_name, "log.md",
                                              prj.results_folder)
            log_file_link = \
                _get_relpath_to_file(log_name, sample_name, prj.results_folder,
                                     get_file_for_project(prj, "reports"))
            log_link_names.append(log_name)
            log_paths.append(log_file_link)
            # get fourth column data (runtime) and fifth column data (memory)
            profile_file_path = \
                _match_file_for_sample(sample.sample_name, 'profile.tsv',
                                       prj.results_folder, full_path=True)
            if os.path.exists(profile_file_path):
                df = _pd.read_csv(profile_file_path, sep="\t", comment="#",
                                  names=PROFILE_COLNAMES)
                df['runtime'] = _pd.to_timedelta(df['runtime'])
                times.append(_get_runtime(df))
                mems.append(_get_maxmem(df))
            else:
                _LOGGER.warning("'{}' does not exist".format(profile_file_path))
                times.append(NO_DATA_PLACEHOLDER)
                mems.append(NO_DATA_PLACEHOLDER)
        else:
            # Sample was not run through the pipeline
            sample_warning.append(sample_name)

    # Alert the user to any warnings generated
    if status_warning:
        _LOGGER.warning("The stats table is incomplete, likely because one or "
                        "more jobs either failed or is still running.")
    if sample_warning:
        _LOGGER.warning("{} samples not present in {}: {}".format(
            len(sample_warning), prj.results_folder,
            str([sample for sample in sample_warning])))
    template_vars = dict(sample_link_names=sample_link_names,
                         row_classes=row_classes, flags=flags, times=times,
                         mems=mems)
    template_name = "status_table_no_links.html"
    if final:
        template_name = "status_table.html"
        template_vars.update(dict(sample_paths=sample_paths,
                                  log_link_names=log_link_names,
                                  log_paths=log_paths))
    return render_jinja_template(template_name, get_jinja_env(), template_vars)


def _get_maxmem(profile_df):
    """
    Get current peak memory

    :param pandas.core.frame.DataFrame profile_df: a data frame representing the current profile.tsv for a sample
    :return str: max memory
    """
    return "{} GB".format(str(max(profile_df['mem']) if not profile_df['mem'].empty else 0))


def _get_runtime(profile_df):
    """
    Collect the unique and last duplicated runtimes, sum them and then return in str format

    :param pandas.core.frame.DataFrame profile_df: a data frame representing the current profile.tsv for a sample
    :return str: sum of runtimes
    """
    unique_df = profile_df[~profile_df.duplicated('cid', keep='last').values]
    return str(timedelta(seconds=sum(unique_df['runtime'].apply(lambda x: x.total_seconds())))).split(".")[0]




""" Pipeline job submission orchestration """

import logging
import os
import subprocess
import time
from jinja2.exceptions import UndefinedError

from attmap import AttMap
from eido import read_schema, validate_inputs
from peppy.const import CONFIG_KEY, SAMPLE_YAML_EXT, SAMPLE_NAME_ATTR

from .processed_project import populate_sample_paths
from .const import *
from .exceptions import JobSubmissionException
from .utils import grab_project_data, fetch_sample_flags, \
    jinja_render_cmd_strictly

_LOGGER = logging.getLogger(__name__)


class SubmissionConductor(object):
    """
    Collects and then submits pipeline jobs.

    This class holds a 'pool' of commands to submit as a single cluster job.
    Eager to submit a job, each instance's collection of commands expands until
    it reaches the 'pool' has been filled, and it's therefore time to submit the
    job. The pool fills as soon as a fill criteria has been reached, which can
    be either total input file size or the number of individual commands.

    """

    def __init__(self, pipeline_interface, prj, delay=0, extra_args=None,
                 extra_args_override=None, ignore_flags=False,
                 compute_variables=None, max_cmds=None, max_size=None,
                 automatic=True, collate=False):
        """
        Create a job submission manager.

        The most critical inputs are the pipeline interface and the pipeline
        key, which together determine which provide critical pipeline
        information like resource allocation packages and which pipeline will
        be overseen by this instance, respectively.

        :param PipelineInterface pipeline_interface: Collection of important
            data for one or more pipelines, like resource allocation packages
            and option/argument specifications
        :param prj: Project with which each sample being considered is
            associated (what generated each sample)
        :param float delay: Time (in seconds) to wait before submitting a job
            once it's ready
        :param str extra_args: string to pass to each job generated,
            for example additional pipeline arguments
        :param str extra_args_override: string to pass to each job generated,
            for example additional pipeline arguments. This deactivates the
            'extra' functionality that appends strings defined in
            Sample.command_extra and Project.looper.command_extra to the
            command template.
        :param bool ignore_flags: Whether to ignore flag files present in
            the sample folder for each sample considered for submission
        :param dict[str] compute_variables: A dict with variables that will be made
            available to the compute package. For example, this should include
            the name of the cluster partition to which job or jobs will be submitted
        :param int | NoneType max_cmds: Upper bound on number of commands to
            include in a single job script.
        :param int | float | NoneType max_size: Upper bound on total file
            size of inputs used by the commands lumped into single job script.
        :param bool automatic: Whether the submission should be automatic once
            the pool reaches capacity.
        :param bool collate: Whether a collate job is to be submitted (runs on
            the project level, rather that on the sample level)
        """

        super(SubmissionConductor, self).__init__()
        self.collate = collate
        self.section_key = PROJECT_PL_KEY if self.collate else SAMPLE_PL_KEY
        self.pl_iface = pipeline_interface
        self.pl_name = self.pl_iface.pipeline_name
        self.prj = prj
        self.compute_variables = compute_variables
        self.extra_pipe_args = extra_args
        self.override_extra = False
        if extra_args_override:
            self.extra_pipe_args = extra_args_override
            self.override_extra = True
        self.ignore_flags = ignore_flags

        self.dry_run = self.prj.dry_run
        self.delay = float(delay)
        self._num_good_job_submissions = 0
        self._num_total_job_submissions = 0
        self._num_cmds_submitted = 0
        self._curr_size = 0
        self._failed_sample_names = []

        if self.extra_pipe_args:
            _LOGGER.debug("String appended to every pipeline command: "
                          "{}".format(self.extra_pipe_args))

        if not self.collate:
            self.automatic = automatic
            if max_cmds is None and max_size is None:
                self.max_cmds = 1
            elif (max_cmds is not None and max_cmds < 1) or \
                    (max_size is not None and max_size < 0):
                raise ValueError(
                    "If specified, max per-job command count must positive, "
                    "and max per-job total file size must be nonnegative")
            else:
                self.max_cmds = max_cmds
            self.max_size = max_size or float("inf")

            self._pool = []
            self._reset_curr_skips()
            self._skipped_sample_pools = []

    @property
    def failed_samples(self):
        return self._failed_sample_names

    @property
    def num_cmd_submissions(self):
        """
        Return the number of commands that this conductor has submitted.
        
        :return int: Number of commands submitted so far.
        """
        return self._num_cmds_submitted

    @property
    def num_job_submissions(self):
        """
        Return the number of jobs that this conductor has submitted.
        
        :return int: Number of jobs submitted so far.
        """
        return self._num_good_job_submissions

    def add_sample(self, sample, rerun=False):
        """
        Add a sample for submission to this conductor.

        :param peppy.Sample sample: sample to be included with this conductor's
            currently growing collection of command submissions
        :param bool rerun: whether the given sample is being rerun rather than
            run for the first time
        :return bool: Indication of whether the given sample was added to
            the current 'pool.'
        :raise TypeError: If sample subtype is provided but does not extend
            the base Sample class, raise a TypeError.
        """
        _LOGGER.debug("Adding {} to conductor for {} to {}run".format(
            sample.sample_name, self.pl_name, "re" if rerun else ""))
        flag_files = fetch_sample_flags(self.prj, sample, self.pl_name)
        use_this_sample = not rerun

        if flag_files or rerun:
            if not self.ignore_flags:
                use_this_sample = False
            # But rescue the sample in case rerun/failed passes
            failed_flag = any("failed" in x for x in flag_files)
            if rerun:
                if failed_flag:
                    _LOGGER.info("> Re-running failed sample")
                    use_this_sample = True
                else:
                    use_this_sample = False
            if not use_this_sample:
                msg = "> Skipping sample"
                if flag_files:
                    msg += ". Flags found: {}".format(flag_files)
                _LOGGER.info(msg)

        if self.prj.toggle_key in sample \
                and int(sample[self.prj.toggle_key]) == 0:
            _LOGGER.warning(
                "> Skiping sample ({}: {})".
                    format(self.prj.toggle_key, sample[self.prj.toggle_key])
            )
            use_this_sample = False

        sample.prj = grab_project_data(self.prj)

        skip_reasons = []
        sample.setdefault("input_file_size", 0)
        # Check for any missing requirements before submitting.
        _LOGGER.debug("Determining missing requirements")
        schema_source = self.pl_iface.get_pipeline_schemas()
        if schema_source and self.prj.file_checks:
            missing = validate_inputs(sample, read_schema(schema_source))
            if missing:
                missing_reqs_msg = "{}: {}".format("Missing files", missing)
                _LOGGER.warning(NOT_SUB_MSG.format(missing_reqs_msg))
                use_this_sample and skip_reasons.append("Missing files")

        if _use_sample(use_this_sample, skip_reasons):
            self._pool.append(sample)
            self._curr_size += float(sample.input_file_size)
            if self.automatic and self._is_full(self._pool, self._curr_size):
                self.submit()
        else:
            self._curr_skip_size += float(sample.input_file_size)
            self._curr_skip_pool.append(sample)
            if self._is_full(self._curr_skip_pool, self._curr_skip_size):
                self._skipped_sample_pools.append((self._curr_skip_pool,
                                                   self._curr_skip_size))
                self._reset_curr_skips()

        return skip_reasons

    def submit(self, force=False):
        """
        Submit one or more commands as a job.
        
        This call will submit the commands corresponding to the current pool 
        of samples if and only if the argument to 'force' evaluates to a 
        true value, or the pool of samples is full.
        
        :param bool force: Whether submission should be done/simulated even
            if this conductor's pool isn't full.
        :return bool: Whether a job was submitted (or would've been if
            not for dry run)
        """
        submitted = False
        if not self._pool:
            _LOGGER.debug("No submission (no pooled samples): %s", self.pl_name)
            # submitted = False
        elif self.collate or force or self._is_full(self._pool, self._curr_size):
            if not self.collate:
                for s in self._pool:
                    schemas = self.prj.get_schemas(self.prj.get_sample_piface(
                        s[SAMPLE_NAME_ATTR]), OUTPUT_SCHEMA_KEY)
                    [populate_sample_paths(s, read_schema(schema))
                     for schema in schemas]
                    s.to_yaml(self._get_sample_yaml_path(s))
            script = self.write_script(self._pool, self._curr_size)
            # Determine whether to actually do the submission.
            _LOGGER.info("Job script (n={0}; {1:.2f}Gb): {2}".
                         format(len(self._pool), self._curr_size, script))
            if self.dry_run:
                _LOGGER.info("Dry run, not submitted")
            elif self._rendered_ok:
                sub_cmd = self.prj.dcc.compute.submission_command
                submission_command = "{} {}".format(sub_cmd, script)
                # Capture submission command return value so that we can
                # intercept and report basic submission failures; #167
                try:
                    subprocess.check_call(submission_command, shell=True)
                except subprocess.CalledProcessError:
                    fails = "" if self.collate \
                        else [s.sample_name for s in self._samples]
                    self._failed_sample_names.extend(fails)
                    self._reset_pool()
                    raise JobSubmissionException(sub_cmd, script)
                time.sleep(self.delay)

            # Update the job and command submission tallies.
            _LOGGER.debug("SUBMITTED")
            if self._rendered_ok:
                submitted = True
                self._num_cmds_submitted += len(self._pool)
            self._reset_pool()

        else:
            _LOGGER.debug("No submission (pool is not full and submission "
                          "was not forced): %s", self.pl_name)
            # submitted = False

        return submitted

    def _get_sample_yaml_path(self, sample):
        """
        Generate path to the sample YAML target location.

        Render path template defined in the pipeline section
        (relative to the pipeline output directory).
        If no template defined, output to the submission directory.

        :param peppy.Sample sample: sample to generate yaml path for
        :return str: path to yaml file
        """
        if SAMPLE_YAML_PATH_KEY not in self.pl_iface:
            return os.path.join(self.prj.submission_folder,
                                "{}{}".format(sample.sample_name,
                                              SAMPLE_YAML_EXT[0]))
        pth_templ = self.pl_iface[SAMPLE_YAML_PATH_KEY]
        namespaces = {"sample": sample,
                      "project": self.prj.prj[CONFIG_KEY],
                      "pipeline": self.pl_iface}
        path = jinja_render_cmd_strictly(pth_templ, namespaces)
        return path if os.path.isabs(path) \
            else os.path.join(self.prj.output_dir, path)

    def _is_full(self, pool, size):
        """
        Determine whether it's time to submit a job for the pool of commands.

        Instances of this class maintain a sort of 'pool' of commands that
        expands as each new command is added, until a time that it's deemed
        'full' and th

        :return bool: Whether this conductor's pool of commands is 'full' and
            ready for submission, as determined by its parameterization
        """
        return self.max_cmds == len(pool) or size >= self.max_size

    @property
    def _samples(self):
        """
        Return a collection of pooled samples.

        :return Iterable[str]: collection of samples currently in the active
            pool for this submission conductor
        """
        return [s for s in self._pool]

    def _sample_lump_name(self, pool):
        """ Determine how to refer to the 'sample' for this submission. """
        if self.collate:
            return "collate"
        if 1 == self.max_cmds:
            assert 1 == len(pool), \
                "If there's a single-command limit on job submission, jobname" \
                " must be determined with exactly one sample in the pool," \
                " but there is/are {}.".format(len(pool))
            sample = pool[0]
            return sample.sample_name
        else:
            # Note the order in which the increment of submission count and
            # the call to this function can influence naming. Make the jobname
            # generation call (this method) before incrementing the
            # submission counter, but add 1 to the index so that we get a
            # name concordant with 1-based, not 0-based indexing.
            return "lump{}".format(self._num_total_job_submissions + 1)

    def _jobname(self, pool):
        """ Create the name for a job submission. """
        return "{}_{}".format(self.pl_iface.pipeline_name,
                              self._sample_lump_name(pool))

    def _set_looper_namespace(self, pool, size):
        """
        Compile a dictionary of looper/submission related settings for use in
        the command templates and in submission script creation
        in divvy (via adapters). Accessible via: {looper.attrname}

        :param Iterable[peppy.Sample] pool: collection of sample instances
        :param float size: cumulative size of the given pool
        :return dict: looper/submission related settings
        """
        settings = AttMap()
        settings.pep_config = self.prj.config_file
        settings.results_subdir = self.prj.results_folder
        settings.submission_subdir = self.prj.submission_folder
        settings.output_dir = self.prj.output_dir
        settings.sample_output_folder = \
            os.path.join(self.prj.results_folder, self._sample_lump_name(pool))
        settings.job_name = self._jobname(pool)
        settings.total_input_size = size
        settings.log_file = \
            os.path.join(self.prj.submission_folder, settings.job_name) + ".log"
        if hasattr(self.prj, "pipeline_config"):
            # Make sure it's a file (it could be provided as null.)
            pl_config_file = self.prj.pipeline_config
            if pl_config_file:
                if not os.path.isfile(pl_config_file):
                    _LOGGER.error("Pipeline config file specified "
                                  "but not found: %s", pl_config_file)
                    raise IOError(pl_config_file)
                _LOGGER.info("Found config file: %s", pl_config_file)
                # Append arg for config file if found
                settings.pipeline_config = pl_config_file
        return settings

    def write_script(self, pool, size):
        """
        Create the script for job submission.

        :param Iterable[peppy.Sample] pool: collection of sample instances
        :param float size: cumulative size of the given pool
        :return str: Path to the job submission script created.
        """
        # looper settings determination
        if self.collate:
            pool = [None]
        looper = self._set_looper_namespace(pool, size)
        commands = []
        namespaces = dict(project=self.prj[CONFIG_KEY],
                          looper=looper,
                          pipeline=self.pl_iface)
        templ = self.pl_iface["command_template"]
        if not self.override_extra:
            extras_template = EXTRA_PROJECT_CMD_TEMPLATE if self.collate \
                else EXTRA_SAMPLE_CMD_TEMPLATE
            templ += extras_template
        for sample in pool:
            # cascading compute settings determination:
            # divcfg < pipeline interface < config <  CLI
            cli = self.compute_variables or {}  # CLI
            if sample:
                namespaces.update({"sample": sample})
            else:
                namespaces.update({"samples": self.prj.samples})
            res_pkg = self.pl_iface.choose_resource_package(namespaces, size or 0)  # config
            res_pkg.update(cli)
            self.prj.dcc.compute.update(res_pkg)  # divcfg
            namespaces.update({"compute": self.prj.dcc.compute})
            self._rendered_ok = False
            try:
                argstring = jinja_render_cmd_strictly(cmd_template=templ,
                                                      namespaces=namespaces)
            except UndefinedError as jinja_exception:
                _LOGGER.warning(NOT_SUB_MSG.format(str(jinja_exception)))
            except KeyError as e:
                exc = "pipeline interface is missing {} section".format(str(e))
                _LOGGER.warning(NOT_SUB_MSG.format(exc))
            else:
                commands.append("{} {}".format(argstring, self.extra_pipe_args))
                self._rendered_ok = True
                self._num_good_job_submissions += 1
                self._num_total_job_submissions += 1
        looper.command = "\n".join(commands)
        if self.collate:
            _LOGGER.debug("samples namespace:\n{}".format(self.prj.samples))
        else:
            _LOGGER.debug("sample namespace:\n{}".format(sample))
        _LOGGER.debug("project namespace:\n{}".format(self.prj[CONFIG_KEY]))
        _LOGGER.debug("pipeline namespace:\n{}".format(self.pl_iface))
        _LOGGER.debug("compute namespace:\n{}".format(self.prj.dcc.compute))
        _LOGGER.debug("looper namespace:\n{}".format(looper))
        subm_base = os.path.join(self.prj.submission_folder, looper.job_name)
        return self.prj.dcc.write_script(output_path=subm_base + ".sub",
                                         extra_vars=[{"looper": looper}])

    def write_skipped_sample_scripts(self):
        """
        For any sample skipped during initial processingwrite submission script
        """
        if self._curr_skip_pool:
            # move any hanging samples from current skip pool to the main pool
            self._skipped_sample_pools.append(
                (self._curr_skip_pool, self._curr_skip_size)
            )
        if self._skipped_sample_pools:
            _LOGGER.info("Writing {} submission scripts for skipped samples".
                          format(len(self._skipped_sample_pools)))
            [self.write_script(pool, size)
             for pool, size in self._skipped_sample_pools]

    def _reset_pool(self):
        """ Reset the state of the pool of samples """
        self._pool = []
        self._curr_size = 0

    def _reset_curr_skips(self):
        self._curr_skip_pool = []
        self._curr_skip_size = 0


def _use_sample(flag, skips):
    return flag and not skips



""" Looper configuration file manager """
import os
from yacman import YacAttMap, select_config
from logging import getLogger
from ubiquerg import expandpath, is_url

from .const import *
from .pipeline_interface import PipelineInterface

_LOGGER = getLogger(__name__)


class LooperConfig(YacAttMap):
    def __init__(self, filepath=None, entries=None):
        """

        :param str filepath:
        :param Mapping entries:
        """
        super(LooperConfig, self).__init__(filepath=filepath, entries=entries)

    def get_pipeline_interface(self, protocol, raw=False):
        """

        :param str protocol:
        :param bool raw:
        :return PipelineInterface: pipeline interface object matched
            by the specified protocol
        """
        if PROTOMAP_KEY in self:
            if protocol in self[PROTOMAP_KEY]:
                return self[PROTOMAP_KEY][protocol] if raw else \
                    PipelineInterface(config=self[PROTOMAP_KEY][protocol])
        return None

    def add_protocol_mapping(self, protocol, loc):
        """

        :param str protocol: protocol key
        :param str loc: path to an existing pipeline interface file
        """
        path = expandpath(loc)
        if not os.path.exists(path):
            if not is_url(loc):
                _LOGGER.warning("Ignoring nonexistent pipeline interface "
                                "location: {}".format(loc))
                return
        else:
            if protocol in self[PROTOMAP_KEY]:
                _LOGGER.info("Overwriting existing protocol mapping with: "
                             "{}:{}".format(protocol, loc))
            self[PROTOMAP_KEY].update({protocol: loc})


def select_looper_config(filename=None, conf_env_vars=CFG_ENV_VARS, **kwargs):
    """
    Get path to looper configuration file.

    :param str filename: name/path of looper configuration file
    :param Iterable[str] conf_env_vars: names of environment
        variables to consider, a prioritized search list
    :return str: path to looper configuration file
    """
    return select_config(filename, conf_env_vars, **kwargs)




"""
Processed Project manipulation functions.
Will be moved to a separate package
"""
import os
from logging import getLogger

from eido.const import *
from eido.exceptions import *

from peppy.sample import Sample
from peppy.project import Project

__author__ = "Michal Stolarczyk"
__email__ = "michal@virginia.edu"

_LOGGER = getLogger(__name__)
PATH_KEY = "path"
THUMB_PATH_KEY = "thumbnail_path"
PATH_LIKE = [PATH_KEY, THUMB_PATH_KEY]


def _get_path_sect_keys(mapping, keys=[PATH_KEY]):
    """
    Get names of subsections in a mapping that contain collection of keys

    :param Mapping mapping: schema subsection to search for paths
    :param  Iterable[str] keys: collection of keys to check for
    :return Iterable[str]: collection of keys to path-like sections
    """
    return [k for k, v in mapping.items() if bool(set(keys) & set(mapping[k]))]


def _populate_paths(object, schema, check_exist):
    """
    Populate path-like object attributes with other object attributes
    based on a defined template, e.g. '/Users/x/test_{name}/{genome}_file.txt'

    :param Mapping object: object with attributes to populate path template with
    :param dict schema: schema with path attributes defined, e.g.
        output of read_schema function
    :param bool check_exist: whether the paths should be check for existence
    :return Mapping: object with path templates populated
    """
    if PROP_KEY not in schema:
        raise EidoSchemaInvalidError("Schema is missing properties section.")
    missing = []
    s = schema[PROP_KEY]
    path_sects = _get_path_sect_keys(s)
    for ps in path_sects:
        templ = s[ps][PATH_KEY]
        try:
            populated = templ.format(**dict(object.items()))
        except Exception as e:
            _LOGGER.warning("Caught exception: {}.\n"
                            "Could not populate path: {}".
                            format(getattr(e, 'message', repr(e)), templ))
        else:
            setattr(object, ps, populated)
            _LOGGER.debug("Path set to: {}".format(object[ps]))
            if check_exist and not os.path.exists(object[ps]):
                missing.append(object[ps])
    if missing:
        raise PathAttrNotFoundError("Path attributes not found:\n- {}".
                                    format("\n- ".join(missing)))


def populate_sample_paths(sample, schema, check_exist=False):
    """
    Populate path-like Sample attributes with other object attributes
    based on a defined template, e.g. '/Users/x/test_{name}/{genome}_file.txt'

    :param peppy.Sample sample: sample to populate paths in
    :param Iterable[dict] schema: schema with path attributes defined, e.g.
        output of read_schema function
    :param bool check_exist: whether the paths should be check for existence
    :return Mapping: Sample with path templates populated
    """
    if not isinstance(sample, Sample):
        raise TypeError("Can only populate paths in peppy.Sample objects")
    # schema = schema[-1]  # use only first schema, in case there are imports
    if PROP_KEY in schema and "samples" in schema[PROP_KEY]:
        _populate_paths(sample, schema[PROP_KEY]["samples"]["items"],
                        check_exist)


def populate_project_paths(project, schema, check_exist=False):
    """
    Populate path-like Project attributes with other object attributes
    based on a defined template, e.g. '/Users/x/test_{name}/{genome}_file.txt'

    :param peppy.Project project: project to populate paths in
    :param dict schema: schema with path attributes defined, e.g.
        output of read_schema function
    :param bool check_exist: whether the paths should be check for existence
    :return Mapping: Project with path templates populated
    """
    if not isinstance(project, Project):
        raise TypeError("Can only populate paths in peppy.Project objects")
    _populate_paths(project, schema, check_exist)


def get_project_outputs(project, schema):
    """
    Get project level outputs with path-like attributes populated with
    project attributes

    :param peppy.Project project:
    :param Iterable[dict] schema:
    :return attmap.PathExAttMap: mapping with populated path-like attributes
    """
    from attmap import PathExAttMap
    # if not any([isinstance(project, Project),
    #             issubclass(type(project), Project)]):
    #     raise TypeError("Can only populate paths in peppy.Project "
    #                     "objects or it subclasses")
    schema = schema[-1]  # use only first schema, in case there are imports
    if PROP_KEY not in schema:
        raise EidoSchemaInvalidError("Schema is missing properties section.")
    res = {}
    s = schema[PROP_KEY]
    path_sects = _get_path_sect_keys(s, keys=PATH_LIKE)
    for ps in path_sects:
        res[ps] = s[ps]
        for p in PATH_LIKE:
            try:
                res[ps][p] = s[ps][p].format(**dict(project.items()))
            except Exception as e:
                _LOGGER.debug("Caught exception: {}.\n Could not populate {} "
                              "path".format(p, str(e)))
    return PathExAttMap(res)



#!/usr/bin/env python
"""
Looper: a pipeline submission engine. https://github.com/pepkit/looper
"""

import abc
import csv
import glob
import logging
import os
import subprocess
import sys
if sys.version_info < (3, 3):
    from collections import Mapping
else:
    from collections.abc import Mapping
import yaml
import pandas as _pd

from collections import defaultdict
# Need specific sequence of actions for colorama imports?
from colorama import init
init()
from colorama import Fore, Style
from shutil import rmtree
from jsonschema import ValidationError
from copy import copy

from . import __version__, build_parser, _LEVEL_BY_VERBOSITY
from .conductor import SubmissionConductor
from .const import *
from .exceptions import JobSubmissionException, MisconfigurationException
from .html_reports import HTMLReportBuilder
from .project import Project, ProjectContext
from .utils import *
from .looper_config import *

from divvy import DEFAULT_COMPUTE_RESOURCES_NAME, select_divvy_config
from logmuse import init_logger
from peppy.const import *
from eido import validate_sample, validate_config, inspect_project
from ubiquerg.cli_tools import query_yes_no
from ubiquerg.collection import uniqify


_PKGNAME = "looper"
_LOGGER = logging.getLogger(_PKGNAME)


class Executor(object):
    """ Base class that ensures the program's Sample counter starts.

    Looper is made up of a series of child classes that each extend the base
    Executor class. Each child class does a particular task (such as run the
    project, summarize the project, destroy the project, etc). The parent
    Executor class simply holds the code that is common to all child classes,
    such as counting samples as the class does its thing."""

    __metaclass__ = abc.ABCMeta

    def __init__(self, prj):
        """
        The Project defines the instance; establish an iteration counter.

        :param Project prj: Project with which to work/operate on
        """
        super(Executor, self).__init__()
        self.prj = prj
        self.counter = LooperCounter(len(prj.samples))

    @abc.abstractmethod
    def __call__(self, *args, **kwargs):
        """ Do the work of the subcommand/program. """
        pass


class Checker(Executor):

    def __call__(self, flags=None, all_folders=False, max_file_count=30):
        """
        Check Project status, based on flag files.

        :param Iterable[str] | str flags: Names of flags to check, optional;
            if unspecified, all known flags will be checked.
        :param bool all_folders: Whether to check flags in all folders, not
            just those for samples in the config file from which the Project
            was created.
        :param int max_file_count: Maximum number of filepaths to display for a
            given flag.
        """

        # Handle single or multiple flags, and alphabetize.
        flags = sorted([flags] if isinstance(flags, str)
                       else list(flags or FLAGS))
        flag_text = ", ".join(flags)

        # Collect the files by flag and sort by flag name.
        _LOGGER.debug("Checking project folders for flags: %s", flag_text)
        if all_folders:
            files_by_flag = fetch_flag_files(
                results_folder=self.prj.results_folder, flags=flags)
        else:
            files_by_flag = fetch_flag_files(prj=self.prj, flags=flags)

        # For each flag, output occurrence count.
        for flag in flags:
            _LOGGER.info("%s: %d", flag.upper(), len(files_by_flag[flag]))

        # For each flag, output filepath(s) if not overly verbose.
        for flag in flags:
            try:
                files = files_by_flag[flag]
            except Exception as e:
                _LOGGER.debug("No files for {} flag. Caught exception: {}".
                              format(flags, getattr(e, 'message', repr(e))))
                continue
            # If checking on a specific flag, do not limit the number of
            # reported filepaths, but do not report empty file lists
            if len(flags) == 1 and len(files) > 0:
                _LOGGER.info("%s (%d):\n%s", flag.upper(),
                             len(files), "\n".join(files))
            # Regardless of whether 0-count flags are previously reported,
            # don't report an empty file list for a flag that's absent.
            # If the flag-to-files mapping is defaultdict, absent flag (key)
            # will fetch an empty collection, so check for length of 0.
            if 0 < len(files) <= max_file_count:
                _LOGGER.info("%s (%d):\n%s", flag.upper(),
                             len(files), "\n".join(files))


class Cleaner(Executor):
    """ Remove all intermediate files (defined by pypiper clean scripts). """

    def __call__(self, args, preview_flag=True):
        """
        Execute the file cleaning process.

        :param argparse.Namespace args: command-line options and arguments
        :param bool preview_flag: whether to halt before actually removing files
        """
        for sample in self.prj.samples:
            _LOGGER.info(self.counter.show(sample.sample_name))
            sample_output_folder = sample_folder(self.prj, sample)
            cleanup_files = glob.glob(os.path.join(sample_output_folder,
                                                   "*_cleanup.sh"))
            if not cleanup_files:
                _LOGGER.info("Nothing to clean.")
                continue
            if preview_flag:
                # Preview: Don't actually clean, just show what will be cleaned.
                _LOGGER.info("Files to clean: %s", ", ".join(cleanup_files))
            else:
                for f in cleanup_files:
                    _LOGGER.info(f)
                    subprocess.call(["sh", f])
        if not preview_flag:
            _LOGGER.info("Clean complete.")
            return 0
        if args.dry_run:
            _LOGGER.info("Dry run. No files cleaned.")
            return 0
        if not args.force_yes and not \
                query_yes_no("Are you sure you want to permanently delete all "
                             "intermediate pipeline results for this project?"):
            _LOGGER.info("Clean action aborted by user.")
            return 1
        self.counter.reset()
        return self(args, preview_flag=False)


class Destroyer(Executor):
    """ Destroyer of files and folders associated with Project's Samples """

    def __call__(self, args, preview_flag=True):
        """
        Completely remove all output produced by any pipelines.

        :param argparse.Namespace args: command-line options and arguments
        :param bool preview_flag: whether to halt before actually removing files
        """

        _LOGGER.info("Removing results:")
        for sample in self.prj.samples:
            _LOGGER.info(self.counter.show(sample.sample_name))
            sample_output_folder = sample_folder(self.prj, sample)
            if preview_flag:
                # Preview: Don't actually delete, just show files.
                _LOGGER.info(str(sample_output_folder))
            else:
                _remove_or_dry_run(sample_output_folder, args.dry_run)

        _LOGGER.info("Removing summary:")
        destroy_summary(self.prj, args.dry_run)

        if not preview_flag:
            _LOGGER.info("Destroy complete.")
            return 0

        if args.dry_run:
            _LOGGER.info("Dry run. No files destroyed.")
            return 0

        if not args.force_yes and not query_yes_no(
                "Are you sure you want to permanently delete all pipeline "
                "results for this project?"):
            _LOGGER.info("Destroy action aborted by user.")
            return 1

        self.counter.reset()

        # Finally, run the true destroy:
        return self(args, preview_flag=False)


class Collator(Executor):
    """" Submitter for project-level pipelines """
    def __init__(self, prj):
        """
        Initializes an instance

        :param Project prj: Project with which to work/operate on
        """
        super(Executor, self).__init__()
        self.prj = prj

    def __call__(self, args, **compute_kwargs):
        """
        Matches collators by protocols, creates submission scripts
        and submits them

        :param argparse.Namespace args: parsed command-line options and
            arguments, recognized by looper
        """
        jobs = 0
        project_pifaces = self.prj.project_pipeline_interface_sources
        if not project_pifaces:
            raise MisconfigurationException(
                "Looper requires a pointer to at least one project pipeline. "
                "Please refer to the documentation on linking project to a "
                "pipeline: "
                "http://looper.databio.org/en/latest/defining-a-project")
        self.counter = LooperCounter(len(project_pifaces))
        for project_piface in project_pifaces:
            try:
                project_piface_object = \
                    PipelineInterface(project_piface, pipeline_type="project")
            except (IOError, ValidationError) as e:
                _LOGGER.warning(
                    "Ignoring invalid pipeline interface source: {}. "
                    "Caught exception: {}".
                        format(project_piface, getattr(e, 'message', repr(e))))
                continue
            _LOGGER.info(self.counter.show(
                name=self.prj.name, type="project",
                pipeline_name=project_piface_object.pipeline_name))
            conductor = SubmissionConductor(
                pipeline_interface=project_piface_object,
                prj=self.prj,
                compute_variables=compute_kwargs,
                delay=args.time_delay,
                extra_args=args.command_extra,
                extra_args_override=args.command_extra_override,
                ignore_flags=args.ignore_flags,
                collate=True
            )
            conductor._pool = [None]
            conductor.submit()
            jobs += conductor.num_job_submissions
        _LOGGER.info("\nLooper finished")
        _LOGGER.info("Jobs submitted: {}".format(jobs))


class Runner(Executor):
    """ The true submitter of pipelines """

    def __call__(self, args, rerun=False, **compute_kwargs):
        """
        Do the Sample submission.

        :param argparse.Namespace args: parsed command-line options and
            arguments, recognized by looper
        :param list remaining_args: command-line options and arguments not
            recognized by looper, germane to samples/pipelines
        :param bool rerun: whether the given sample is being rerun rather than
            run for the first time
        """
        max_cmds = sum(list(map(len, self.prj._samples_by_interface.values())))
        self.counter.total = max_cmds
        failures = defaultdict(list)  # Collect problems by sample.
        processed_samples = set()  # Enforce one-time processing.
        submission_conductors = {}
        try:
            comp_vars = self.prj.dcc[COMPUTE_KEY].to_map()
        except AttributeError:
            if not isinstance(self.prj.dcc[COMPUTE_KEY], Mapping):
                raise TypeError("Project's computing config isn't a mapping: {}"
                                " ({})".format(self.prj.dcc[COMPUTE_KEY],
                                               type(self.prj.dcc[COMPUTE_KEY])))
            from copy import deepcopy
            comp_vars = deepcopy(self.prj.dcc[COMPUTE_KEY])
        comp_vars.update(compute_kwargs or {})

        # Determine number of samples eligible for processing.
        num_samples = len(self.prj.samples)
        if args.limit is None:
            upper_sample_bound = num_samples
        elif args.limit < 0:
            raise ValueError("Invalid number of samples to run: {}".format(args.limit))
        else:
            upper_sample_bound = min(args.limit, num_samples)
        _LOGGER.debug("Limiting to {} of {} samples".
                      format(upper_sample_bound, num_samples))

        num_commands_possible = 0
        failed_submission_scripts = []

        # config validation (samples excluded) against all schemas defined
        # for every pipeline matched for this project
        [validate_config(self.prj, schema_file, True)
         for schema_file in self.prj.get_schemas(self.prj.pipeline_interfaces)]

        for piface in self.prj.pipeline_interfaces:
            conductor = SubmissionConductor(
                pipeline_interface=piface,
                prj=self.prj,
                compute_variables=comp_vars,
                delay=args.time_delay,
                extra_args=args.command_extra,
                extra_args_override=args.command_extra_override,
                ignore_flags=args.ignore_flags,
                max_cmds=args.lumpn,
                max_size=args.lump
            )
            submission_conductors[piface.pipe_iface_file] = conductor

        for sample in self.prj.samples[:upper_sample_bound]:
            pl_fails = []
            skip_reasons = []
            sample_pifaces = self.prj.get_sample_piface(sample[SAMPLE_NAME_ATTR])
            if not sample_pifaces:
                skip_reasons.append("No pipeline interfaces defined")

            if skip_reasons:
                _LOGGER.warning(NOT_SUB_MSG.format(", ".join(skip_reasons)))
                failures[sample.sample_name] = skip_reasons
                continue

            # single sample validation against a single schema
            # (from sample's piface)
            [validate_sample(self.prj, sample.sample_name, schema_file, True)
             for schema_file in self.prj.get_schemas(sample_pifaces)]

            processed_samples.add(sample[SAMPLE_NAME_ATTR])

            for sample_piface in sample_pifaces:
                _LOGGER.info(
                    self.counter.show(name=sample.sample_name,
                                      pipeline_name=sample_piface.pipeline_name)
                )
                num_commands_possible += 1
                cndtr = submission_conductors[sample_piface.pipe_iface_file]
                try:
                    curr_pl_fails = cndtr.add_sample(sample, rerun=rerun)
                except JobSubmissionException as e:
                    failed_submission_scripts.append(e.script)
                else:
                    pl_fails.extend(curr_pl_fails)
            if pl_fails:
                failures[sample.sample_name].extend(pl_fails)

        job_sub_total = 0
        cmd_sub_total = 0

        for piface, conductor in submission_conductors.items():
            conductor.submit(force=True)
            job_sub_total += conductor.num_job_submissions
            cmd_sub_total += conductor.num_cmd_submissions
            conductor.write_skipped_sample_scripts()

        # Report what went down.
        _LOGGER.info("\nLooper finished")
        _LOGGER.info("Samples valid for job generation: {} of {}".
                     format(len(processed_samples), num_samples))
        _LOGGER.info("Commands submitted: {} of {}".
                     format(cmd_sub_total, max_cmds))
        _LOGGER.info("Jobs submitted: {}".format(job_sub_total))
        if args.dry_run:
            _LOGGER.info("Dry run. No jobs were actually submitted.")

        # Restructure sample/failure data for display.
        samples_by_reason = defaultdict(set)
        # Collect names of failed sample(s) by failure reason.
        for sample, failures in failures.items():
            for f in failures:
                samples_by_reason[f].add(sample)
        # Collect samples by pipeline with submission failure.
        for piface, conductor in submission_conductors.items():
            # Don't add failure key if there are no samples that failed for
            # that reason.
            if conductor.failed_samples:
                fails = set(conductor.failed_samples)
                samples_by_reason[SUBMISSION_FAILURE_MESSAGE] |= fails

        failed_sub_samples = samples_by_reason.get(SUBMISSION_FAILURE_MESSAGE)
        if failed_sub_samples:
            _LOGGER.info("\n{} samples with at least one failed job submission:"
                         " {}".format(len(failed_sub_samples),
                                      ", ".join(failed_sub_samples)))

        # If failure keys are only added when there's at least one sample that
        # failed for that reason, we can display information conditionally,
        # depending on whether there's actually failure(s).
        if samples_by_reason:
            _LOGGER.info("\n{} unique reasons for submission failure: {}".format(
                len(samples_by_reason), ", ".join(samples_by_reason.keys())))
            full_fail_msgs = [_create_failure_message(reason, samples)
                              for reason, samples in samples_by_reason.items()]
            _LOGGER.info("\nSummary of failures:\n{}".
                         format("\n".join(full_fail_msgs)))


class Report(Executor):
    """ Combine project outputs into a browsable HTML report """
    def __call__(self, args):
        # initialize the report builder
        report_builder = HTMLReportBuilder(self.prj)

        # Do the stats and object summarization.
        table = Table(self.prj)()
        # run the report builder. a set of HTML pages is produced
        report_path = report_builder(table.objs, table.stats,
                                     uniqify(table.columns))

        _LOGGER.info("HTML Report (n=" + str(len(table.stats)) + "): "
                     + report_path)


class Table(Executor):
    """ Project/Sample statistics and table output generator """
    def __init__(self, prj):
        # call the inherited initialization
        super(Table, self).__init__(prj)
        self.prj = prj

    def __call__(self):
        # pull together all the fits and stats from each sample into
        # project-combined spreadsheets.
        self.stats, self.columns = _create_stats_summary(self.prj, self.counter)
        self.objs = _create_obj_summary(self.prj, self.counter)
        return self


def _create_stats_summary(project, counter):
    """
    Create stats spreadsheet and columns to be considered in the report, save
    the spreadsheet to file

    :param looper.Project project: the project to be summarized
    :param looper.LooperCounter counter: a counter object
    """
    # Create stats_summary file
    columns = []
    stats = []
    project_samples = project.samples
    missing_files = []
    _LOGGER.info("Creating stats summary...")
    for sample in project_samples:
        _LOGGER.info(counter.show(sample.sample_name, sample.protocol))
        sample_output_folder = sample_folder(project, sample)
        # Grab the basic info from the annotation sheet for this sample.
        # This will correspond to a row in the output.
        sample_stats = sample.get_sheet_dict()
        columns.extend(sample_stats.keys())
        # Version 0.3 standardized all stats into a single file
        stats_file = os.path.join(sample_output_folder, "stats.tsv")
        if not os.path.isfile(stats_file):
            missing_files.append(stats_file)
            continue
        t = _pd.read_csv(stats_file, sep="\t", header=None,
                         names=['key', 'value', 'pl'])
        t.drop_duplicates(subset=['key', 'pl'], keep='last', inplace=True)
        t.loc[:, 'plkey'] = t['pl'] + ":" + t['key']
        dupes = t.duplicated(subset=['key'], keep=False)
        t.loc[dupes, 'key'] = t.loc[dupes, 'plkey']
        sample_stats.update(t.set_index('key')['value'].to_dict())
        stats.append(sample_stats)
        columns.extend(t.key.tolist())
    if missing_files:
        _LOGGER.warning("Stats files missing for {} samples: {}".
                        format(len(missing_files),missing_files))
    tsv_outfile_path = get_file_for_project(project, 'stats_summary.tsv')
    tsv_outfile = open(tsv_outfile_path, 'w')
    tsv_writer = csv.DictWriter(tsv_outfile, fieldnames=uniqify(columns),
                                delimiter='\t', extrasaction='ignore')
    tsv_writer.writeheader()
    for row in stats:
        tsv_writer.writerow(row)
    tsv_outfile.close()
    _LOGGER.info("Statistics summary (n=" + str(len(stats)) + "): " +
                 tsv_outfile_path)
    counter.reset()
    return stats, uniqify(columns)


def _create_obj_summary(project, counter):
    """
    Read sample specific objects files and save to a data frame

    :param looper.Project project: the project to be summarized
    :param looper.LooperCounter counter: a counter object
    :return pandas.DataFrame: objects spreadsheet
    """
    _LOGGER.info("Creating objects summary...")
    objs = _pd.DataFrame()
    # Create objects summary file
    missing_files = []
    for sample in project.samples:
        # Process any reported objects
        _LOGGER.info(counter.show(sample.sample_name, sample.protocol))
        sample_output_folder = sample_folder(project, sample)
        objs_file = os.path.join(sample_output_folder, "objects.tsv")
        if not os.path.isfile(objs_file):
            missing_files.append(objs_file)
            continue
        t = _pd.read_csv(objs_file, sep="\t", header=None,
                         names=['key', 'filename', 'anchor_text',
                                'anchor_image', 'annotation'])
        t['sample_name'] = sample.sample_name
        objs = objs.append(t, ignore_index=True)
    if missing_files:
        _LOGGER.warning("Object files missing for {} samples: {}".
                        format(len(missing_files), missing_files))
    # create the path to save the objects file in
    objs_file = get_file_for_project(project, 'objs_summary.tsv')
    objs.to_csv(objs_file, sep="\t")
    _LOGGER.info("Objects summary (n=" +
                 str(len(project.samples) - len(missing_files)) + "): " +
                 objs_file)
    return objs


def _create_failure_message(reason, samples):
    """ Explain lack of submission for a single reason, 1 or more samples. """
    color = Fore.LIGHTRED_EX
    reason_text = color + reason + Style.RESET_ALL
    samples_text = ", ".join(samples)
    return "{}: {}".format(reason_text, samples_text)


def _remove_or_dry_run(paths, dry_run=False):
    """
    Remove file or directory or just inform what would be removed in
    case of dry run

    :param list|str paths: list of paths to files/dirs to be removed
    :param bool dry_run: logical indicating whether the files should remain
        untouched and massage printed
    """
    paths = paths if isinstance(paths, list) else [paths]
    for path in paths:
        if os.path.exists(path):
            if dry_run:
                _LOGGER.info("DRY RUN. I would have removed: " + path)
            else:
                _LOGGER.info("Removing: " + path)
                if os.path.isfile(path):
                    os.remove(path)
                else:
                    rmtree(path, ignore_errors=True)
        else:
            _LOGGER.info(path + " does not exist.")


def destroy_summary(prj, dry_run=False):
    """
    Delete the summary files if not in dry run mode
    """
    _remove_or_dry_run([get_file_for_project(prj, "summary.html"),
                        get_file_for_project(prj, 'stats_summary.tsv'),
                        get_file_for_project(prj, 'objs_summary.tsv'),
                        get_file_for_project(prj, "reports")], dry_run)


class LooperCounter(object):
    """
    Count samples as you loop through them, and create text for the
    subcommand logging status messages.

    :param int total: number of jobs to process
    """

    def __init__(self, total):
        self.count = 0
        self.total = total

    def show(self, name, type="sample", pipeline_name=None):
        """
        Display sample counts status for a particular protocol type.

        The counts are running vs. total for the protocol within the Project,
        and as a side-effect of the call, the running count is incremented.

        :param str name: name of the sample
        :param str pipeline_name: name of the pipeline
        :return str: message suitable for logging a status update
        """
        self.count += 1
        return _submission_status_text(type=type,
            curr=self.count, total=self.total, name=name,
            pipeline_name=pipeline_name, color=Fore.CYAN
        )

    def reset(self):
        self.count = 0

    def __str__(self):
        return "LooperCounter of size {}".format(self.total)


def _submission_status_text(curr, total, name, pipeline_name=None,
                            type="sample", color=Fore.CYAN):
    """ Generate submission sample text for run or collate """
    txt = color + "## [{n} of {t}] {type}: {name}".\
        format(n=curr, t=total, type=type, name=name)
    if pipeline_name:
        txt += "; pipeline: {}".format(pipeline_name)
    return txt + Style.RESET_ALL


def _proc_resources_spec(args):
    """
    Process CLI-sources compute setting specification. There are two sources
    of compute settings in the CLI alone:
        * YAML file (--settings argument)
        * itemized compute settings (--compute argument)

    The itemized compute specification is given priority

    :param argparse.Namespace: arguments namespace
    :return Mapping[str, str]: binding between resource setting name and value
    :raise ValueError: if interpretation of the given specification as encoding
        of key-value pairs fails
    """
    spec = getattr(args, "compute", None)
    try:
        settings_data = read_yaml_file(args.settings) or {}
    except yaml.YAMLError:
        _LOGGER.warning("Settings file ({}) does not follow YAML format,"
                        " disregarding".format(args.settings))
        settings_data = {}
    if not spec:
        return settings_data
    pairs = [(kv, kv.split("=")) for kv in spec]
    bads = []
    for orig, pair in pairs:
        try:
            k, v = pair
        except ValueError:
            bads.append(orig)
        else:
            settings_data[k] = v
    if bads:
        raise ValueError(
            "Could not correctly parse itemized compute specification. "
            "Correct format: " + EXAMPLE_COMPUTE_SPEC_FMT)
    return settings_data


def main():
    """ Primary workflow """
    global _LOGGER
    parsers = build_parser()
    parser = parsers[0]
    aux_parser = parsers[1]
    aux_parser.suppress_defaults()
    args, remaining_args = parser.parse_known_args()
    if args.command is None:
        parser.print_help(sys.stderr)
        sys.exit(1)
    if args.config_file is None:
        m = "No project config defined"
        try:
            setattr(args, "config_file", read_cfg_from_dotfile())
        except OSError:
            print(m + " and dotfile does not exist: {}".format(dotfile_path()))
            parser.print_help(sys.stderr)
            sys.exit(1)
        else:
            print(m + ", using: {}. Read from dotfile ({}).".
                  format(read_cfg_from_dotfile(), dotfile_path()))
    if args.command == "init":
        sys.exit(int(not init_dotfile(dotfile_path(), args.config_file, args.force)))
    args = enrich_args_via_cfg(args, aux_parser)

    # Set the logging level.
    if args.dbg:
        # Debug mode takes precedence and will listen for all messages.
        level = args.logging_level or logging.DEBUG
    elif args.verbosity is not None:
        # Verbosity-framed specification trumps logging_level.
        level = _LEVEL_BY_VERBOSITY[args.verbosity]
    else:
        # Normally, we're not in debug mode, and there's not verbosity.
        level = LOGGING_LEVEL

    # Establish the project-root logger and attach one for this module.
    logger_kwargs = {"level": level,
                     "logfile": args.logfile,
                     "devmode": args.dbg}
    init_logger(name="peppy", **logger_kwargs)
    init_logger(name="divvy", **logger_kwargs)
    init_logger(name="eido", **logger_kwargs)
    _LOGGER = init_logger(name=_PKGNAME, **logger_kwargs)

    # lc = LooperConfig(select_looper_config(filename=args.looper_config))
    # _LOGGER.debug("Determined genome config: {}".format(lc))

    _LOGGER.info("Looper version: {}\nCommand: {}".
                 format(__version__, args.command))

    if len(remaining_args) > 0:
        _LOGGER.warning("Unrecognized arguments: {}".
                      format(" ".join([str(x) for x in remaining_args])))

    divcfg = select_divvy_config(filepath=args.divvy) \
        if hasattr(args, "divvy") else None

    # Initialize project
    _LOGGER.debug("Building Project")
    try:
        p = Project(config_file=args.config_file,
                    amendments=args.amend,
                    divcfg_path=divcfg,
                    runp=args.command == "runp",
                    **{attr: getattr(args, attr) for attr in CLI_PROJ_ATTRS if attr in args})
    except yaml.parser.ParserError as e:
        _LOGGER.error("Project config parse failed -- {}".format(e))
        sys.exit(1)

    selected_compute_pkg = p.selected_compute_package \
                           or DEFAULT_COMPUTE_RESOURCES_NAME
    if p.dcc is not None and not p.dcc.activate_package(selected_compute_pkg):
        _LOGGER.info("Failed to activate '{}' computing package. "
                     "Using the default one".format(selected_compute_pkg))

    with ProjectContext(prj=p,
                        selector_attribute=args.sel_attr,
                        selector_include=args.sel_incl,
                        selector_exclude=args.sel_excl) as prj:

        if args.command in ["run", "rerun"]:
            run = Runner(prj)
            try:
                compute_kwargs = _proc_resources_spec(args)
                run(args, rerun=(args.command == "rerun"), **compute_kwargs)
            except IOError:
                _LOGGER.error("{} pipeline_interfaces: '{}'".
                              format(prj.__class__.__name__,
                                     prj.pipeline_interface_sources))
                raise

        if args.command == "runp":
            compute_kwargs = _proc_resources_spec(args)
            collate = Collator(prj)
            collate(args, **compute_kwargs)

        if args.command == "destroy":
            return Destroyer(prj)(args)

        if args.command == "table":
            Table(prj)()
        
        if args.command == "report":
            Report(prj)(args)

        if args.command == "check":
            Checker(prj)(flags=args.flags)

        if args.command == "clean":
            return Cleaner(prj)(args)

        if args.command == "inspect":
            inspect_project(p, args.snames, args.attr_limit)



""" Utility functions for internal, developmental use """

import copy
from logmuse import init_logger

__author__ = "Vince Reuter"
__email__ = "vreuter@virginia.edu"

__all__ = ["est_log"]


def est_log(**kwargs):
    """
    Establish logging, e.g. for an interactive session.

    :param dict kwargs: keyword arguments for logger setup.
    :return logging.Logger: looper logger
    """
    kwds = copy.copy(kwargs)
    if "name" in kwds:
        print("Ignoring {} and setting fixed values for logging names".
              format(kwds["name"]))
        del kwds["name"]
    init_logger(name="peppy", **kwds)
    return init_logger(name="looper", **kwds)



""" Helpers without an obvious logical home. """

from collections import defaultdict, Iterable
from logging import getLogger
import glob
import os
from .const import *
from .exceptions import MisconfigurationException
from peppy.const import *
from peppy import Project as peppyProject
import jinja2
import yaml
import argparse
from ubiquerg import convert_value, expandpath

_LOGGER = getLogger(__name__)


def fetch_flag_files(prj=None, results_folder="", flags=FLAGS):
    """
    Find all flag file paths for the given project.

    :param Project | AttributeDict prj: full Project or AttributeDict with
        similar metadata and access/usage pattern
    :param str results_folder: path to results folder, corresponding to the
        1:1 sample:folder notion that a looper Project has. That is, this
        function uses the assumption that if results_folder rather than project
        is provided, the structure of the file tree rooted at results_folder is
        such that any flag files to be found are not directly within rootdir but
        are directly within on of its first layer of subfolders.
    :param Iterable[str] | str flags: Collection of flag names or single flag
        name for which to fetch files
    :return Mapping[str, list[str]]: collection of filepaths associated with
        particular flag for samples within the given project
    :raise TypeError: if neither or both of project and rootdir are given
    """

    if not (prj or results_folder) or (prj and results_folder):
        raise TypeError("Need EITHER project OR rootdir")

    # Just create the filenames once, and pair once with flag name.
    flags = [flags] if isinstance(flags, str) else list(flags)
    flagfile_suffices = ["*{}.flag".format(f) for f in flags]
    flag_suffix_pairs = list(zip(flags, flagfile_suffices))

    # Collect the flag file paths by flag name.
    files_by_flag = defaultdict(list)

    if prj is None:
        for flag, suffix in flag_suffix_pairs:
            flag_expr = os.path.join(results_folder, "*", suffix)
            flags_present = glob.glob(flag_expr)
            files_by_flag[flag] = flags_present
    else:
        # Iterate over samples to collect flag files.
        for s in prj.samples:
            folder = sample_folder(prj, s)
            # Check each candidate flag for existence, collecting if present.
            for flag, suffix in flag_suffix_pairs:
                flag_expr = os.path.join(folder, suffix)
                flags_present = glob.glob(flag_expr)
                files_by_flag[flag].extend(flags_present)

    return files_by_flag


def fetch_sample_flags(prj, sample, pl_name):
    """
    Find any flag files present for a sample associated with a project

    :param looper.Project prj: project of interest
    :param peppy.Sample sample: sample object of interest
    :param str pl_name: name of the pipeline for which flag(s) should be found
    :return Iterable[str]: collection of flag file path(s) associated with the
        given sample for the given project
    """
    sfolder = sample_folder(prj=prj, sample=sample)
    if not os.path.isdir(sfolder):
        _LOGGER.debug("Results folder ({}) doesn't exist for sample {}".
                      format(sfolder, str(sample)))
        return []
    folder_contents = [os.path.join(sfolder, f) for f in os.listdir(sfolder)]
    return [x for x in folder_contents if os.path.splitext(x)[1] == ".flag"
            and os.path.basename(x).startswith(pl_name)]


def grab_project_data(prj):
    """
    From the given Project, grab Sample-independent data.

    There are some aspects of a Project of which it's beneficial for a Sample
    to be aware, particularly for post-hoc analysis. Since Sample objects
    within a Project are mutually independent, though, each doesn't need to
    know about any of the others. A Project manages its, Sample instances,
    so for each Sample knowledge of Project data is limited. This method
    facilitates adoption of that conceptual model.

    :param Project prj: Project from which to grab data
    :return Mapping: Sample-independent data sections from given Project
    """
    if not prj:
        return {}

    try:
        data = prj[CONFIG_KEY]
    except KeyError:
        _LOGGER.debug("Project lacks section '%s', skipping", CONFIG_KEY)
    return data


def sample_folder(prj, sample):
    """
    Get the path to this Project's root folder for the given Sample.

    :param AttributeDict | Project prj: project with which sample is associated
    :param Mapping sample: Sample or sample data for which to get root output
        folder path.
    :return str: this Project's root folder for the given Sample
    """
    return os.path.join(prj.results_folder,
                        sample[SAMPLE_NAME_ATTR])


def get_file_for_project(prj, appendix):
    """
    Create a path to the file for the current project.
    Takes the possibility of amendment being activated at the time

    :param looper.Project prj: project object
    :param str appendix: the appendix of the file to create the path for,
        like 'objs_summary.tsv' for objects summary file
    :return str: path to the file
    """
    fp = os.path.join(prj.output_dir, prj[NAME_KEY])
    if hasattr(prj, AMENDMENTS_KEY) and getattr(prj, AMENDMENTS_KEY):
        fp += '_' + '_'.join(getattr(prj, AMENDMENTS_KEY))
    fp += '_' + appendix
    return fp


def jinja_render_cmd_strictly(cmd_template, namespaces):
    """
    Render a command string in the provided namespaces context.

    Strictly, which means that all the requested attributes must be
    available in the namespaces

    :param str cmd_template: command template do be filled in with the
        variables in the provided namespaces. For example:
        "prog.py --name {project.name} --len {sample.len}"
    :param Mapping[Mapping[str] namespaces: context for command rendering.
        Possible namespaces are: looper, project, sample, pipeline
    :return str: rendered command
    """
    def _finfun(x):
        """
        A callable that can be used to process the result of a variable
        expression before it is output. Joins list elements
        """
        return " ".join(x) if isinstance(x, list) else x

    env = jinja2.Environment(undefined=jinja2.StrictUndefined,
                             variable_start_string="{",
                             variable_end_string="}",
                             finalize=_finfun)
    template = env.from_string(cmd_template)
    try:
        rendered = template.render(**namespaces)
    except jinja2.exceptions.UndefinedError:
        _LOGGER.error("Missing sample, project or pipeline attributes"
                      " required by command template: '{}'"
                      .format(cmd_template))
        raise
    _LOGGER.debug("rendered arg str: {}".format(rendered))
    return rendered


def read_yaml_file(filepath):
    """
    Read a YAML file

    :param str filepath: path to the file to read
    :return dict: read data
    """
    data = None
    if os.path.exists(filepath):
        with open(filepath, 'r') as f:
            data = yaml.safe_load(f)
    return data


def enrich_args_via_cfg(parser_args, aux_parser):
    """
    Read in a looper dotfile and set arguments.

    Priority order: CLI > dotfile/config > parser default

    :param argparse.Namespace parser_args: parsed args by the original parser
    :param argparse.Namespace aux_parser: parsed args by the a parser
        with defaults suppressed
    :return argparse.Namespace: selected argument values
    """
    cfg_args_all = \
        _get_subcommand_args(parser_args) \
            if os.path.exists(parser_args.config_file) else dict()
    result = argparse.Namespace()
    cli_args, _ = aux_parser.parse_known_args()
    for dest in vars(parser_args):
        if dest not in POSITIONAL or not hasattr(result, dest):
            if dest in cli_args:
                x = getattr(cli_args, dest)
                r = convert_value(x) if isinstance(x, str) else x
            elif dest in cfg_args_all:
                if isinstance(cfg_args_all[dest], list):
                    r = [convert_value(i) for i in cfg_args_all[dest]]
                else:
                    r = convert_value(cfg_args_all[dest])
            else:
                r = getattr(parser_args, dest)
            setattr(result, dest, r)
    return result


def _get_subcommand_args(parser_args):
    """
    Get the union of values for the subcommand arguments from
    Project.looper, Project.looper.cli.<subcommand> and Project.looper.cli.all.
    If any are duplicated, the above is the selection priority order.

    Additionally, convert the options strings to destinations (replace '-'
    with '_'), which strongly relies on argument parser using default
    destinations.

    :param argparser.Namespace parser_args: argument namespace
    :return dict: mapping of argument destinations to their values
    """
    args = dict()
    cfg = peppyProject(parser_args.config_file,
                       defer_samples_creation=True,
                       amendments=parser_args.amend)
    if CONFIG_KEY in cfg and LOOPER_KEY in cfg[CONFIG_KEY] \
            and CLI_KEY in cfg[CONFIG_KEY][LOOPER_KEY]:
        try:
            cfg_args = cfg[CONFIG_KEY][LOOPER_KEY][CLI_KEY] or dict()
            args = cfg_args[ALL_SUBCMD_KEY] or dict() \
                if ALL_SUBCMD_KEY in cfg_args else dict()
            args.update(cfg_args[parser_args.command] or dict()
                        if parser_args.command in cfg_args else dict())
        except (TypeError, KeyError, AttributeError, ValueError) as e:
            raise MisconfigurationException(
                "Invalid '{}.{}' section in the config. Caught exception: {}".
                    format(LOOPER_KEY, CLI_KEY, getattr(e, 'message', repr(e))))
    if CONFIG_KEY in cfg and LOOPER_KEY in cfg[CONFIG_KEY]:
        try:
            if CLI_KEY in cfg[CONFIG_KEY][LOOPER_KEY]:
                del cfg[CONFIG_KEY][LOOPER_KEY][CLI_KEY]
            args.update(cfg[CONFIG_KEY][LOOPER_KEY])
        except (TypeError, KeyError, AttributeError, ValueError) as e:
            raise MisconfigurationException(
                "Invalid '{}' section in the config. Caught exception: {}".
                    format(LOOPER_KEY, getattr(e, 'message', repr(e))))
    args = {k.replace("-", "_"): v for k, v in args.items()} if args else None
    return args


def init_dotfile(path, cfg_path, force=False):
    """
    Initialize looper dotfile

    :param str path: absolute path to the file to initialize
    :param str cfg_path: path to the config file. Absolute or relative to 'path'
    :param bool force: whether the existing file should be overwritten
    :return bool: whether the file was initialized
    """
    if os.path.exists(path) and not force:
        print("Can't initialize, file exists: {}".format(path))
        return False
    cfg_path = expandpath(cfg_path)
    if not os.path.isabs(cfg_path):
        cfg_path = os.path.join(os.path.dirname(path), cfg_path)
    assert os.path.exists(cfg_path), \
        OSError("Provided config path is invalid. You must provide path "
                "that is either absolute or relative to: {}".
                format(os.path.dirname(path)))
    relpath = os.path.relpath(cfg_path, os.path.dirname(path))
    with open(path, 'w') as dotfile:
        yaml.dump({DOTFILE_CFG_PTH_KEY: relpath}, dotfile)
    print("Initialized looper dotfile: {}".format(path))
    return True


def read_cfg_from_dotfile():
    """
    Read file path to the config file from the dotfile

    :return str: path to the config file read from the dotfile
    :raise MisconfigurationException: if the dotfile does not consist of the
        required key pointing to the PEP
    """
    dp = dotfile_path(must_exist=True)
    with open(dp, 'r') as dotfile:
        dp_data = yaml.safe_load(dotfile)
    if DOTFILE_CFG_PTH_KEY in dp_data:
        return os.path.join(os.path.dirname(dp),
                            str(os.path.join(dp_data[DOTFILE_CFG_PTH_KEY])))
    else:
        raise MisconfigurationException(
            "Looper dotfile ({}) is missing '{}' key".
                format(dp, DOTFILE_CFG_PTH_KEY))


def dotfile_path(directory=os.getcwd(), must_exist=False):
    """
    Get the path to the looper dotfile

    If file existence is forced this function will look for it in
    the directory parents

    :param str directory: directory path to start the search in
    :param bool must_exist: whether the file must exist
    :return str: path to the dotfile
    :raise OSError: if the file does not exist
    """
    cur_dir = directory
    if not must_exist:
        return os.path.join(cur_dir, LOOPER_DOTFILE_NAME)
    while True:
        parent_dir = os.path.dirname(cur_dir)
        if LOOPER_DOTFILE_NAME in os.listdir(cur_dir):
            return os.path.join(cur_dir, LOOPER_DOTFILE_NAME)
        if cur_dir == parent_dir:
            # root, file does not exist
            raise OSError("Looper dotfile ({}) not found in '{}' and all "
                          "its parents".format(LOOPER_DOTFILE_NAME, directory))
        cur_dir = parent_dir



#! /usr/bin/env python

import os
from setuptools import setup
import sys


# Additional keyword arguments for setup().
extra = {}


# Ordinary dependencies
DEPENDENCIES = []
with open("requirements/requirements-all.txt", "r") as reqs_file:
    for line in reqs_file:
        if not line.strip():
            continue
        #DEPENDENCIES.append(line.split("=")[0].rstrip("<>"))
        DEPENDENCIES.append(line)


# numexpr for pandas
try:
    import numexpr
except ImportError:
    # No numexpr is OK for pandas.
    pass
else:
    # pandas 0.20.2 needs updated numexpr; the claim is 2.4.6, but that failed.
    DEPENDENCIES.append("numexpr>=2.6.2")
extra["install_requires"] = DEPENDENCIES


# 2to3
if sys.version_info >= (3, ):
    extra["use_2to3"] = True


# Additional files to include with package
def get_static(name, condition=None):
    static = [os.path.join(name, f) for f in os.listdir(
            os.path.join(os.path.dirname(os.path.realpath(__file__)), name))]
    if condition is None:
        return static
    else:
        return [i for i in filter(lambda x: eval(condition), static)]


# scripts to be added to the $PATH
# scripts = get_static("scripts", condition="'.' in x")
# scripts removed (TO remove this)
scripts = None


with open("looper/_version.py", 'r') as versionfile:
    version = versionfile.readline().split()[-1].strip("\"'\n")

# Handle the pypi README formatting.
try:
    import pypandoc
    long_description = pypandoc.convert_file('README.md', 'rst')
except(IOError, ImportError, OSError):
    long_description = open('README.md').read()

setup(
    name="looper",
    packages=["looper"],
    version=version,
    description="A pipeline submission engine that parses sample inputs and submits pipelines for each sample.",
    long_description=long_description,
    long_description_content_type='text/markdown', 
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Scientific/Engineering :: Bio-Informatics"
    ],
    keywords="bioinformatics, sequencing, ngs",
    url="https://github.com/pepkit/looper",
    author=u"Nathan Sheffield, Vince Reuter, Michal Stolarczyk, Johanna Klughammer, Andre Rendeiro",
    license="BSD2",
    entry_points={
        "console_scripts": [
            'looper = looper.__main__:main'
        ],
    },
    scripts=scripts,
    package_data={'looper': ['submit_templates/*']},
    include_package_data=True,
    test_suite="tests",
    tests_require=(["mock", "pytest"]),
    setup_requires=(["pytest-runner"] if {"test", "pytest", "ptr"} & set(sys.argv) else []),
    **extra
)



"""Project configuration, particularly for logging.

Project-scope constants may reside here, but more importantly, some setup here
will provide a logging infrastructure for all of the project's modules.
Individual modules and classes may provide separate configuration on a more
local level, but this will at least provide a foundation.

"""

import argparse
import os
import logging
from .conductor import SubmissionConductor
from .pipeline_interface import PipelineInterface
from .project import Project
from ._version import __version__
from .parser_types import *
from .const import *

from ubiquerg import VersionInHelpParser
from divvy import DEFAULT_COMPUTE_RESOURCES_NAME, NEW_COMPUTE_KEY as COMPUTE_KEY
# Not used here, but make this the main import interface between peppy and
# looper, so that other modules within this package need not worry about
# the locations of some of the peppy declarations. Effectively, concentrate
# the connection between peppy and looper here, to the extent possible.

__all__ = ["Project", "PipelineInterface", "SubmissionConductor"]

# Descending by severity for correspondence with logic inversion.
# That is, greater verbosity setting corresponds to lower logging level.
_LEVEL_BY_VERBOSITY = [logging.ERROR, logging.CRITICAL, logging.WARN,
                       logging.INFO, logging.DEBUG]


class _StoreBoolActionType(argparse.Action):
    """
    Enables the storage of a boolean const and custom type definition needed
    for systematic html interface generation. To get the _StoreTrueAction
    output use default=False in the add_argument function
    and default=True to get _StoreFalseAction output.
    """
    def __init__(self, option_strings, dest, type, default,
                 required=False, help=None):
        super(_StoreBoolActionType, self).__init__(
            option_strings=option_strings,
            dest=dest,
            nargs=0,
            const=not default,
            default=default,
            type=type,
            required=required,
            help=help)

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, self.const)

        
def build_parser():
    """
    Building argument parser.

    :return argparse.ArgumentParser
    """
    # Main looper program help text messages
    banner = "%(prog)s - A project job submission engine and project manager."
    additional_description = "For subcommand-specific options, " \
                             "type: '%(prog)s <subcommand> -h'"
    additional_description += "\nhttps://github.com/pepkit/looper"

    parser = VersionInHelpParser(
        prog="looper", description=banner, epilog=additional_description,
        version=__version__)

    aux_parser = VersionInHelpParser(
        prog="looper", description=banner, epilog=additional_description,
        version=__version__)
    result = []
    for parser in [parser, aux_parser]:
        # Logging control
        parser.add_argument(
                "--logfile", help="Optional output file for looper logs "
                                  "(default: %(default)s)")
        parser.add_argument(
                "--verbosity", type=int, choices=range(len(_LEVEL_BY_VERBOSITY)),
                help="Choose level of verbosity (default: %(default)s)")
        parser.add_argument(
                "--logging-level", help=argparse.SUPPRESS)
        parser.add_argument(
                "--dbg", action="store_true",
                help="Turn on debug mode (default: %(default)s)")
        # Individual subcommands
        msg_by_cmd = {
                "run": "Run or submit sample jobs.",
                "rerun": "Resubmit sample jobs with failed flags.",
                "runp": "Run or submit project jobs.",
                "table": "Write summary stats table for project samples.",
                "report": "Create browsable HTML report of project results.",
                "destroy": "Remove output files of the project.",
                "check": "Check flag status of current runs.",
                "clean": "Run clean scripts of already processed jobs.",
                "inspect": "Print information about a project.",
                "init": "Initialize looper dotfile."
        }

        subparsers = parser.add_subparsers(dest="command")

        def add_subparser(cmd):
            message = msg_by_cmd[cmd]
            return subparsers.add_parser(cmd, description=message, help=message,
                formatter_class=lambda prog: argparse.HelpFormatter(
                    prog, max_help_position=37, width=90))

        # Run and rerun command
        run_subparser = add_subparser("run")
        rerun_subparser = add_subparser("rerun")
        collate_subparser = add_subparser("runp")
        table_subparser = add_subparser("table")
        report_subparser = add_subparser("report")
        destroy_subparser = add_subparser("destroy")
        check_subparser = add_subparser("check")
        clean_subparser = add_subparser("clean")
        inspect_subparser = add_subparser("inspect")
        init_subparser = add_subparser("init")

        # Flag arguments
        ####################################################################
        for subparser in [run_subparser, rerun_subparser, collate_subparser]:
            subparser.add_argument(
                    "-i", "--ignore-flags", default=False,
                    action=_StoreBoolActionType, type=html_checkbox(checked=False),
                    help="Ignore run status flags? Default=False")

        for subparser in [run_subparser, rerun_subparser, destroy_subparser,
                          clean_subparser, collate_subparser]:
            subparser.add_argument(
                    "-d", "--dry-run",
                    action=_StoreBoolActionType, default=False,
                    type=html_checkbox(checked=False),
                    help="Don't actually submit the jobs.  Default=False")

        # Parameter arguments
        ####################################################################
        for subparser in [run_subparser, rerun_subparser, collate_subparser]:
            subparser.add_argument(
                    "-t", "--time-delay", metavar="S",
                    type=html_range(min_val=0, max_val=30, value=0), default=0,
                    help="Time delay in seconds between job submissions")
            subparser.add_argument(
                    "-l", "--limit", default=None, metavar="N",
                    type=html_range(min_val=1, max_val="num_samples",
                                    value="num_samples"),
                    help="Limit to n samples")
            subparser.add_argument(
                    "-x", "--command-extra", default="",
                    metavar="S", help="String to append to every command")
            subparser.add_argument(
                    "-y", "--command-extra-override", metavar="S", default="",
                    help="Same as command-extra, but overrides values in PEP")
            subparser.add_argument(
                    "-f", "--skip-file-checks",
                    action=_StoreBoolActionType, default=False,
                    type=html_checkbox(checked=False),
                    help="Do not perform input file checks")

            divvy_group = \
                subparser.add_argument_group(
                    "divvy arguments",
                    "Configure divvy to change computing settings")
            divvy_group.add_argument(
                "--divvy", default=None, metavar="DIVCFG",
                help="Path to divvy configuration file. Default=$DIVCFG env "
                     "variable. Currently: {}".format(os.getenv('DIVCFG', None)
                                                      or "not set"))
            divvy_group.add_argument(
                    "-p", "--package", metavar="P",
                    help="Name of computing resource package to use")
            divvy_group.add_argument(
                    "-s", "--settings", default="", metavar="S",
                    help="Path to a YAML settings file with compute settings")
            divvy_group.add_argument(
                    "-c", "--compute", metavar="K", nargs="+",
                    help="List of key-value pairs (k1=v1)")

        for subparser in [run_subparser, rerun_subparser]:
            subparser.add_argument(
                    "-u", "--lump", default=None, metavar="X",
                    type=html_range(min_val=0, max_val=100, step=0.1, value=0),
                    help="Total input file size (GB) to batch into one job")
            subparser.add_argument(
                    "-n", "--lumpn", default=None, metavar="N",
                    type=html_range(min_val=1, max_val="num_samples", value=1),
                    help="Number of commands to batch into one job")

        inspect_subparser.add_argument(
            "-n", "--snames", required=False, nargs="+", metavar="S",
            help="Name of the samples to inspect")
        inspect_subparser.add_argument(
            "-l", "--attr-limit", required=False, type=int, default=10,
            metavar="L", help="Number of sample attributes to display")

        check_subparser.add_argument(
                "-A", "--all-folders", action=_StoreBoolActionType,
                default=False, type=html_checkbox(checked=False),
                help="Check status for all  output folders, not just for "
                     "samples specified in the config. Default=False")
        check_subparser.add_argument(
                "-f", "--flags", nargs='*', default=FLAGS,
                type=html_select(choices=FLAGS), metavar="F",
                help="Check on only these flags/status values")

        for subparser in [destroy_subparser, clean_subparser]:
            subparser.add_argument(
                    "--force-yes", action=_StoreBoolActionType, default=False,
                    type=html_checkbox(checked=False),
                    help="Provide upfront confirmation of destruction intent, "
                         "to skip console query.  Default=False")

        init_subparser.add_argument("config_file", help="Project configuration "
                                                        "file (YAML)")

        init_subparser.add_argument("-f", "--force", help="Force overwrite",
            action="store_true", default=False)

        # Common arguments
        for subparser in [run_subparser, rerun_subparser, table_subparser,
                          report_subparser, destroy_subparser, check_subparser,
                          clean_subparser, collate_subparser, inspect_subparser]:
            subparser.add_argument("config_file", nargs="?", default=None,
                                   help="Project configuration file (YAML)")
            # help="Path to the output directory"
            subparser.add_argument("-o", "--output-dir", metavar="DIR",
                                   help=argparse.SUPPRESS)
            # "Submission subdirectory name"
            subparser.add_argument("--submission-subdir", metavar="DIR",
                                   help=argparse.SUPPRESS)
            # "Results subdirectory name"
            subparser.add_argument("--results-subdir", metavar="DIR",
                                   help=argparse.SUPPRESS)
            # "Sample attribute for pipeline interface sources"
            subparser.add_argument("--pipeline-interfaces-key", metavar="K",
                                   help=argparse.SUPPRESS)
            # "Paths to pipeline interface files"
            subparser.add_argument("--pipeline-interfaces", metavar="P",
                                   nargs="+", action="append",
                                   help=argparse.SUPPRESS)

        for subparser in [run_subparser, rerun_subparser, table_subparser,
                          report_subparser, destroy_subparser, check_subparser,
                          clean_subparser, collate_subparser, inspect_subparser]:
            fetch_samples_group = \
                subparser.add_argument_group(
                    "sample selection arguments",
                    "Specify samples to include or exclude based on sample attribute values")
            fetch_samples_group.add_argument(
                "-g", "--toggle-key", metavar="K",
                help="Sample attribute specifying toggle. Default: toggle")
            fetch_samples_group.add_argument(
                "--sel-attr", default="toggle", metavar="ATTR",
                help="Attribute for sample exclusion OR inclusion")
            protocols = fetch_samples_group.add_mutually_exclusive_group()
            protocols.add_argument(
                    "--sel-excl", nargs='*', metavar="E",
                    help="Exclude samples with these values")
            protocols.add_argument(
                    "--sel-incl", nargs='*', metavar="I",
                    help="Include only samples with these values")
            subparser.add_argument(
                    "-a", "--amend", nargs="+", metavar="A",
                    help="List of amendments to activate")
        result.append(parser)
    return result



""" Shared project constants """

import os

__author__ = "Databio lab"
__email__ = "nathan@code.databio.org"

__all__ = [
    "BUTTON_APPEARANCE_BY_FLAG", "TABLE_APPEARANCE_BY_FLAG",
    "ID_COLNAME", "NO_DATA_PLACEHOLDER", "OUTKEY", "ALL_SUBCMD_KEY",
    "OUTDIR_KEY", "LOOPER_KEY", "COMPUTE_KEY", "PIPELINE_INTERFACES_KEY",
    "SIZE_DEP_VARS_KEY", "FLAGS", "DYN_VARS_KEY", "SAMPLE_YAML_PATH_KEY",
    "RESOURCES_KEY", "NOT_SUB_MSG", "EXTRA_KEY", "DEFAULT_CFG_PATH",
    "PIFACE_SCHEMA_SRC", "RESULTS_SUBDIR_KEY", "SUBMISSION_SUBDIR_KEY",
    "TEMPLATES_DIRNAME", "FILE_SIZE_COLNAME", "COMPUTE_PACKAGE_KEY",
    "INPUT_SCHEMA_KEY", "OUTPUT_SCHEMA_KEY", "EXAMPLE_COMPUTE_SPEC_FMT",
    "SAMPLE_PL_KEY", "PROJECT_PL_KEY", "CFG_ENV_VARS", "LOGGING_LEVEL",
    "PIFACE_KEY_SELECTOR", "SUBMISSION_FAILURE_MESSAGE", "IMAGE_EXTS",
    "PROFILE_COLNAMES", "SAMPLE_TOGGLE_ATTR", "TOGGLE_KEY_SELECTOR",
    "LOOPER_DOTFILE_NAME", "POSITIONAL", "EXTRA_PROJECT_CMD_TEMPLATE",
    "EXTRA_SAMPLE_CMD_TEMPLATE", "SELECTED_COMPUTE_PKG", "CLI_PROJ_ATTRS",
    "DOTFILE_CFG_PTH_KEY", "DRY_RUN_KEY", "FILE_CHECKS_KEY", "CLI_KEY"
]

FLAGS = ["completed", "running", "failed", "waiting", "partial"]

APPEARANCE_BY_FLAG = {
    "completed": {
        "button_class": "{type}-success",
        "flag": "Completed"
    },
    "running": {
        "button_class": "{type}-primary",
        "flag": "Running"
    },
    "failed": {
        "button_class": "{type}-danger",
        "flag": "Failed"
    },
    "parital": {
        "button_class": "{type}-warning",
        "flag": "Partial"
    },
    "waiting": {
        "button_class": "{type}-info",
        "flag": "Waiting"
    }
}


def _get_apperance_dict(type, templ=APPEARANCE_BY_FLAG):
    """
    Based on the type of the HTML element provided construct the appearence
     mapping using the template

    :param dict templ: appearance template to populate
    :param str type: type of HTML element to populate template with
    :return dict: populated appearance template
    """
    from copy import deepcopy
    ret = deepcopy(templ)
    for flag, app_dict in ret.items():
        for key, app in app_dict.items():
            ret[flag][key] = ret[flag][key].format(type=type)
    return ret


LOGGING_LEVEL = "INFO"
CFG_ENV_VARS = ["LOOPER"]
TABLE_APPEARANCE_BY_FLAG = _get_apperance_dict("table")
BUTTON_APPEARANCE_BY_FLAG = _get_apperance_dict("btn btn")
NO_DATA_PLACEHOLDER = "NA"
PIFACE_KEY_SELECTOR = "pipeline_interfaces_key"
PIPELINE_INTERFACES_KEY = "pipeline_interfaces"
RESOURCES_KEY = "resources"
SAMPLE_PL_KEY = "sample_pipeline"
PROJECT_PL_KEY = "project_pipeline"
PIFACE_SCHEMA_SRC = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                 "schemas", "pipeline_interface_schema_{}.yaml")
EXTRA_SAMPLE_CMD_TEMPLATE = "{%- if sample.command_extra is defined %} {sample.command_extra}   {% endif -%}"
EXTRA_PROJECT_CMD_TEMPLATE = "{%- if project.looper.command_extra is defined %} {project.looper.command_extra}{% endif -%}"
DOTFILE_CFG_PTH_KEY = "config_file_path"
INPUT_SCHEMA_KEY = "input_schema"
OUTPUT_SCHEMA_KEY = "output_schema"
SAMPLE_YAML_PATH_KEY = "sample_yaml_path"
TOGGLE_KEY_SELECTOR = "toggle_key"
SAMPLE_TOGGLE_ATTR = "toggle"
OUTKEY = "outputs"
COMPUTE_KEY = "compute"
COMPUTE_PACKAGE_KEY = "package"
SIZE_DEP_VARS_KEY = "size_dependent_variables"
DYN_VARS_KEY = "dynamic_variables_command_template"
TEMPLATES_DIRNAME = "jinja_templates"
NOT_SUB_MSG = "> Not submitted: {}"
IMAGE_EXTS = ('.png', '.jpg', '.jpeg', '.svg', '.gif')
PROFILE_COLNAMES = ['pid', 'hash', 'cid', 'runtime', 'mem', 'cmd', 'lock']  # this strongly depends on pypiper's profile.tsv format

PIPE_ARGS_SECTION = "pipeline_args"
CLI_KEY = "cli"
LOOPER_KEY = "looper"
OUTDIR_KEY = "output_dir"
RESULTS_SUBDIR_KEY = "results_subdir"
SUBMISSION_SUBDIR_KEY = "submission_subdir"
DRY_RUN_KEY = "dry_run"
FILE_CHECKS_KEY = "skip_file_checks"
EXAMPLE_COMPUTE_SPEC_FMT = "k1=v1 k2=v2"
SUBMISSION_FAILURE_MESSAGE = "Cluster resource failure"
LOOPER_DOTFILE_NAME = "." + LOOPER_KEY + ".yaml"
POSITIONAL = ["config_file", "command"]
SELECTED_COMPUTE_PKG = "package"
EXTRA_KEY = "_cli_extra"
ALL_SUBCMD_KEY = "all"
DEFAULT_CFG_PATH = os.path.join(os.getcwd(), LOOPER_DOTFILE_NAME)
CLI_PROJ_ATTRS = [OUTDIR_KEY, TOGGLE_KEY_SELECTOR, SUBMISSION_SUBDIR_KEY, PIPELINE_INTERFACES_KEY,
                  RESULTS_SUBDIR_KEY, PIFACE_KEY_SELECTOR, COMPUTE_PACKAGE_KEY, DRY_RUN_KEY, FILE_CHECKS_KEY]

# resource package TSV-related consts
ID_COLNAME = "id"
FILE_SIZE_COLNAME = "max_file_size"
IMAGE_EXTS = ('.png', '.jpg', '.jpeg', '.svg', '.gif')
PROFILE_COLNAMES = ['pid', 'hash', 'cid', 'runtime', 'mem', 'cmd', 'lock']  # this strongly depends on pypiper's profile.tsv format



""" Exceptions for specific looper issues. """

from abc import ABCMeta
from collections import Iterable

__author__ = "Vince Reuter"
__email__ = "vreuter@virginia.edu"

_all__ = ["DuplicatePipelineKeyException",
          "InvalidResourceSpecificationException",
          "JobSubmissionException",
          "LooperError",
          "MissingPipelineConfigurationException",
          "PipelineInterfaceConfigError",
          "PipelineInterfaceRequirementsError",
          "MisconfigurationException"]


class LooperError(Exception):
    """ Base type for custom package errors. """
    __metaclass__ = ABCMeta


class MisconfigurationException(LooperError):
    """ Duplication of pipeline identifier precludes unique pipeline ref. """
    def __init__(self, key):
        super(MisconfigurationException, self).__init__(key)


class DuplicatePipelineKeyException(LooperError):
    """ Duplication of pipeline identifier precludes unique pipeline ref. """
    def __init__(self, key):
        super(DuplicatePipelineKeyException, self).__init__(key)


class InvalidResourceSpecificationException(LooperError):
    """ Pipeline interface resources--if present--needs default. """
    def __init__(self, reason):
        super(InvalidResourceSpecificationException, self).__init__(reason)


class JobSubmissionException(LooperError):
    """ Error type for when job submission fails. """
    def __init__(self, sub_cmd, script):
        self.script = script
        reason = "Error for command {} and script '{}'".\
            format(sub_cmd, self.script)
        super(JobSubmissionException, self).__init__(reason)


class MissingPipelineConfigurationException(LooperError):
    """ A selected pipeline needs configuration data. """
    def __init__(self, pipeline):
        super(MissingPipelineConfigurationException, self).__init__(pipeline)


class PipelineInterfaceConfigError(LooperError):
    """ Error with PipelineInterface config data during construction. """
    def __init__(self, context):
        """
        For exception context, provide message or collection of missing sections.

        :param str | Iterable[str] context:
        """
        if not isinstance(context, str) and isinstance(context, Iterable):
            context = "Missing section(s): {}".format(", ".join(context))
        super(PipelineInterfaceConfigError, self).__init__(context)


class PipelineInterfaceRequirementsError(LooperError):
    """ Invalid specification of pipeline requirements in interface config. """
    def __init__(self, typename_by_requirement):
        super(PipelineInterfaceRequirementsError, self).__init__(
            "{} invalid requirements: {}".format(
                len(typename_by_requirement), typename_by_requirement))
        self.error_specs = typename_by_requirement
